#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
import aiohttp
import asyncio
from enum import Enum
import json
import math
from nltk.corpus import stopwords as nltk_stopwords
from nltk.tokenize import word_tokenize
from nltk.stem import PorterStemmer
import os
import random
import string
import time
from typing import List, Tuple, Optional, Dict, Any, Set, Union

from parlai.agents.ir_baseline.ir_baseline import score_match, MaxPriorityQueue
from parlai.core.dict import DictionaryAgent
from parlai.core.message import Message
from parlai.core.metrics import F1Metric
from parlai.core.torch_generator_agent import PPLMetric
import parlai.utils.logging as logging
import projects.bb3.constants as CONST
import projects.bb3.prompts as PROMPT
from projects.bb3.agents.module import Module

try:
    import contractions
except ImportError as e:
    print('please install contractions: "pip install contractions"')
    raise e

STEMMER = PorterStemmer()

# Very limited number of stopwords:
# We don't remove you and me etc. because of their strong importance in memory records.
_STOP_WORDS = set(
    list(string.punctuation)
    + ['a', 'an', 'the', 'am', 'are', 'is', 'as']
    + [
        STEMMER.stem(adv)
        for adv in (
            'really',
            'actually',
            'up',
            'so',
            'just',
            'now',
            'how',
            'then',
            'also',
            'very',
            'too',
            'most',
            'but',
        )
    ]
)


_STOP_WORDS_WITH_PRONOUNS = set(nltk_stopwords.words('english')).union(_STOP_WORDS)


def normal_tokenizer(
    text: str,
    remove_contractions: bool = True,
    stem=True,
    include_pronouns: bool = False,
) -> List[str]:
    """
    Returns normalized tokens for `text`
    """
    assert isinstance(text, str), 'The only valid arg type is str.'
    text = text.strip().lower()
    if not text:
        return []

    if remove_contractions:
        text = contractions.fix(text)

    tokens = word_tokenize(text)

    if stem:
        tokens = [STEMMER.stem(t) for t in tokens]
    stop_words = _STOP_WORDS if not include_pronouns else _STOP_WORDS_WITH_PRONOUNS
    tokens = [t for t in tokens if t not in stop_words]

    return tokens


def no_overlap(
    data_source: List[str],
    canidate_text: str,
    non_overlap_threshold: int = 1,
    remove_contractions: bool = True,
) -> bool:
    """
    Returns True if tokenized `canidate_text` text has at least `non_overlap_threshold`
    tokens NOT overlapping with all texts in `data_source` list.

    We use this for de-duplicating (for example memory entries)
    """
    tokenized_candidate_text = set(
        normal_tokenizer(
            canidate_text,
            remove_contractions=remove_contractions,
        )
    )
    for dst in data_source:
        dst_tokens = set(
            normal_tokenizer(
                dst,
                remove_contractions=remove_contractions,
            )
        )
        n_overlap = len(dst_tokens - tokenized_candidate_text)
        logging.debug(
            f'"{dst}" AND "{canidate_text}" ---- non-overlapping tokens: {n_overlap}'
        )
        if n_overlap < non_overlap_threshold:
            logging.info(
                f'"{canidate_text}" has too much overlap with "{dst}": discarding repeated entry.'
            )
            return False

    return True


def clean_text(text: str) -> str:
    """
    Removes all special tokens from an incoming text.
    """
    for token in CONST.ALL_SPECIAL_TOKENS:
        text = text.replace(f" {token}", '')
        text = text.replace(f"{token} ", '')
        text = text.replace(token, '')

    return text


class Decision(Enum):

    ALWAYS = 'always'
    NEVER = 'never'
    COMPUTE = 'compute'


def is_opener(text: str, mems: Optional[Dict[str, int]]):
    """
    Return if message is an opener!
    """
    return (
        text == PROMPT.OPENING_PREFIX
        and mems is not None
        and (isinstance(mems, list) or isinstance(mems, dict))
        and len(mems) > 0
    )


def set_failed_reply(reply: Message) -> Message:
    """
    Set proper fields in reply if failed.
    """
    reply.force_set('id', 'FallBackRoutine')
    reply.force_set('text', APIUtils.METASEQ_FAIL_MESSAGE_TEXT)
    reply.force_set('score', float('inf'))
    return reply


################
# Memory Utils #
################


class MemoryUtils:
    self_memory_prefix: str = PROMPT.SELF_MEMORY_PREFIX
    partner_memory_prefix: str = PROMPT.PARTNER_MEMORY_PREFIX

    self_prefix: str = PROMPT.SELF_PREFIX
    partner_prefix: str = PROMPT.PARTNER_PREFIX

    @classmethod
    def is_opt_ft_mem_format(cls, memory_text: str) -> bool:
        return (
            ':' in memory_text
            and memory_text.startswith(cls.self_memory_prefix)
            or memory_text.startswith(cls.partner_memory_prefix)
        )

    @classmethod
    def _is_r2c2_format(cls, memory_text: str) -> bool:
        return (
            ':' in memory_text
            and memory_text.startswith("your persona")
            or memory_text.startswith("partner's persona")
        )

    @classmethod
    def _is_opt_prompt_format(cls, memory_text: str) -> bool:
        return memory_text.startswith(cls.self_prefix) or memory_text.startswith(
            cls.partner_prefix
        )

    @classmethod
    def validate_memory_format(cls, memory_text: str):
        assert (
            MemoryUtils._is_r2c2_format(memory_text)
            or MemoryUtils._is_opt_prompt_format(memory_text)
            or MemoryUtils.is_opt_ft_mem_format(memory_text)
        ), f'Provided memory "{memory_text}" has invalid format for chatbot memory field.'

    @classmethod
    def split_prefix_memory(cls, memory_text: str) -> Tuple[str, str]:
        MemoryUtils.validate_memory_format(memory_text)
        try:
            prfx, mem = memory_text.split(':', 1)
        except ValueError:
            # prompt agent sometimes says things like,
            # "Person 2 is"...
            assert memory_text.startswith(cls.self_prefix) or memory_text.startswith(
                cls.partner_prefix
            )
            if memory_text.startswith(cls.self_prefix):
                prfx, mem = (
                    memory_text[: len(cls.self_prefix)],
                    memory_text[len(cls.self_prefix) + 1 :],
                )
            else:
                prfx, mem = (
                    memory_text[: len(cls.partner_prefix)],
                    memory_text[len(cls.partner_prefix) + 1 :],
                )
        return prfx.strip(), mem.strip()

    @classmethod
    def is_valid_memory(
        cls,
        chatbot_memory: Union[List[str], Dict[str, int]],
        new_memory: str,
        new_memory_prefix: str,
    ) -> bool:
        """
        Return whether the memory is valid.

        It rejects new memory entry as invalid if one similar to it exists already.
        """
        if new_memory in (
            CONST.NOPERSONA,
            PROMPT.NO_MEMORY,
            '',
            APIUtils.METASEQ_FAIL_MESSAGE_TEXT,
        ):
            return False

        # Rejecting if identical memories exist already.
        if not chatbot_memory:
            # No need to dedup if there is no existing memory
            return True

        # filtering for the memories that applies to the current person (self or other)
        person_memories = [
            MemoryUtils.split_prefix_memory(mem)[1]
            for mem in chatbot_memory
            if MemoryUtils.split_prefix_memory(mem)[0] == new_memory_prefix
        ]
        if not person_memories:
            # No memory on record for this person
            return True

        return no_overlap(person_memories, new_memory)

    @classmethod
    def get_memory_prefix(cls, self_or_partner: str, model_type: str) -> str:
        """
        Return memory prefix.
        """
        assert self_or_partner in ['self', 'partner']
        assert model_type in ['R2C2', 'OPT']
        self_prefix = 'your persona' if model_type == 'R2C2' else cls.self_memory_prefix
        partner_prefix = (
            "partner's persona" if model_type == 'R2C2' else cls.partner_memory_prefix
        )
        if self_or_partner == 'self':
            return self_prefix
        else:
            return partner_prefix

    @classmethod
    def add_memory_prefix(
        cls, memory: str, self_or_partner: str, model_type: str
    ) -> str:
        """
        Ensure that the memory has a "persona" prefix.

        :param memory:
            memory to prefix
        :param self_or_partner:
            string representing if this is a self memory or partner memory
        :param model_type:
            which model we're working with

        :return prefixed_mem:
            return a memory with the appropriate prefix.
        """
        assert self_or_partner in ['self', 'partner']
        assert model_type in ['R2C2', 'OPT']
        prefix = MemoryUtils.get_memory_prefix(self_or_partner, model_type)
        if model_type == 'R2C2':
            if not memory.startswith(prefix):
                memory = f"{prefix}: {memory}"
        elif self_or_partner == 'self':
            memory = memory.replace("Person 1", "Person 2")
            if not memory.startswith('Person'):
                memory = f"{prefix}: {memory}"
        else:
            memory = memory.replace("Person 2", "Person 1")
            if not memory.startswith('Person'):
                memory = f"{prefix}: {memory}"

        return memory

    @classmethod
    def maybe_add_memory_prefix(
        cls, memory: str, self_or_partner: str, model_type: str
    ) -> str:
        """
        Maybe add it if it's not there.
        """
        if not MemoryUtils.is_opt_ft_mem_format(memory):
            memory = MemoryUtils.add_memory_prefix(memory, self_or_partner, model_type)
        return memory

    @classmethod
    def _build_query_representation(
        cls, query: str, dictionary: DictionaryAgent
    ) -> Dict[str, Any]:
        rep = {}
        rep['words'] = {}
        words = [w for w in dictionary.tokenize(query.lower())]
        rw = rep['words']
        used = {}
        for w in words:
            rw[w] = 1.0 / (1.0 + math.log(1.0 + dictionary.freq[w]))
            used[w] = True
        rep['norm'] = math.sqrt(len(words))
        return rep

    @classmethod
    def maybe_reduce_memories(
        cls, text: str, memories: List[str], dictionary: DictionaryAgent
    ) -> List[str]:
        """
        TFIDF-Match memories with the textual input to reduce num memories.

        :param observation:
            raw observation
        :param memories:
            memories from which to choose

        :return memories:
            return - potentially shortened - list of memories
        """
        new_memories = []
        if (
            not memories or len(memories) < 32
        ):  # 512 / 16, assuming 16 tokens max per memory
            return memories
        mpq = MaxPriorityQueue(1000)
        query = MemoryUtils._build_query_representation(text, dictionary)
        for m in memories:
            score = score_match(query, m, 0, dictionary)
            mpq.add(m, score)
        new_memories = list(reversed(mpq))[:32]
        return new_memories

    @classmethod
    def get_available_memories(
        cls,
        text: str,
        memories: Dict[str, int],
        in_session_memories: Set[str],
        dictionary: Optional[DictionaryAgent] = None,
        ignore_in_session_memories: bool = False,
        memory_overlap_threshold: float = 0.0,
        memory_hard_block_for_n_turns: int = 0,
        memory_soft_block_decay_factor: float = 0.0,
    ) -> List[str]:
        """
        Return available memories.

        :param text:
            incoming partner text
        :param memories:
            list of all memories
        :param in_session_memories:
            set of memories generated within the current conversation session
        :param ignore_in_session_memories:
            whether to ignore memories generated within the session
        """
        available_memory = []
        for memory, turns_since_used in memories.items():
            turns_since_used = int(turns_since_used)
            # check if we should ignore in session memories
            if ignore_in_session_memories and memory in in_session_memories:
                continue
            # check overlap
            if memory_overlap_threshold > 0:
                non_stopword_memory = ' '.join(
                    normal_tokenizer(memory.split(':')[-1], include_pronouns=True)
                )
                non_stopword_text = ' '.join(
                    normal_tokenizer(text, include_pronouns=True)
                )
                if (
                    F1Metric.compute(non_stopword_memory, [non_stopword_text]).value()
                    < memory_overlap_threshold
                ):
                    continue
            # check hard block
            if turns_since_used < memory_hard_block_for_n_turns:
                continue
            # check soft block
            if memory_soft_block_decay_factor > 0 and random.random() < (
                memory_soft_block_decay_factor**turns_since_used
            ):
                continue

            available_memory.append(memory)
        return MemoryUtils.maybe_reduce_memories(text, available_memory, dictionary)

    @classmethod
    def add_memory(cls, memory: str, memories: Dict[str, int]) -> Dict[str, int]:
        """
        Add memory to the memory store.

        :param memory:
            memory to add
        :param memories:
            all the memories

        :return memories:
            return memories with new memory
        """
        if not memory:
            return memories
        assert memory not in memories
        memories[memory] = 0
        return memories

    @classmethod
    def update_memory_usage(
        cls, used_memory: str, memories: Dict[str, int]
    ) -> Dict[str, int]:
        """
        Update memories to indicate that a memory was used.

        :param memory:
            used memory
        :param memories:
            all the memories

        :return memories:
            return memories with usage updated
        """
        assert not used_memory or used_memory in memories
        for mem in memories:
            if mem == used_memory:
                memories[mem] = 0
            else:
                memories[mem] += 1
        return memories


#################
# OPT API UTILS #
#################


class APIUtils:

    DEFAULT_KEY = os.environ.get("USER", "parlai")
    DEFAULT_SERVER = "DEFAULT_API_SERVER:6010"
    DEFAULT_SERVER_TIMEOUT = 600
    METASEQ_FAIL_MESSAGE_TEXT = 'METASEQ RESPONSE FAILED.'

    @staticmethod
    def is_request_failed_response(resp):
        """
        Whether the requests to Metaseq worker have failed.

        It checks this based on the existences of the failure reasons as they get
        accumulated in `_make_request` functionn calls.
        """
        return len(
            resp.get('failures', [])
        ) > 0 or APIUtils.METASEQ_FAIL_MESSAGE_TEXT in resp.get('text', '')

    @staticmethod
    async def make_request(
        session,
        server: str,
        api_key: str,
        prompt: str,
        min_tokens: int = 0,
        n: int = 1,
        max_tokens: int = 32,
        best_of: int = 1,
        top_p: float = -1,
        echo: bool = False,
        stop: Optional[str] = None,
        temperature: float = 1.0,
        num_retry_on_api_exception=-1,
        lambda_decay: float = -1,
        omega_bound: float = 0.3,
        request_delay: float = 0.5,
        alpha_presence: float = 0.0,
        alpha_frequency: float = 0.0,
        alpha_presence_src: float = 0.0,
        alpha_frequency_src: float = 0.0,
        alpha_src_penalty_end_idx: int = -1,
    ):
        data = {
            'prompt': prompt,
            'min_tokens': min_tokens,
            'max_tokens': max_tokens,
            'best_of': best_of,
            'top_p': top_p,
            'stop': stop,
            'temperature': temperature,
            'echo': echo,
            'lambda_decay': lambda_decay,
            'omega_bound': omega_bound,
            'alpha_presence': alpha_presence,
            'alpha_frequency': alpha_frequency,
            'alpha_presence_src': alpha_presence_src,
            'alpha_frequency_src': alpha_frequency_src,
            "alpha_src_penalty_end_idx": alpha_src_penalty_end_idx,
        }
        init_request_delay = request_delay
        past_exceptions = []
        while True:
            if (
                num_retry_on_api_exception >= 0
                and len(past_exceptions) > num_retry_on_api_exception
            ):
                logging.error('Reached maximum retries, returning failure message.')
                return {
                    'failures': past_exceptions,
                }
            try:
                logging.debug(f'Making request: {data}')
                headers = {'Authorization': f'Bearer {api_key}'}
                async with session.post(
                    f'{server}/completions', json=data, headers=headers
                ) as resp:
                    resp_text = await resp.text()
                    obj = json.loads(resp_text)
                    if 'error' in obj:
                        request_delay *= 2
                        logging.warning(f"Error: {obj['error']}")
                        past_exceptions.append(f"API Error: {obj['error']}")
                        logging.debug(past_exceptions[-1])
                        continue
                    debug = json.dumps(obj, sort_keys=True)
                    logging.debug(f'GPT-Z response: {debug}')
                    request_delay = init_request_delay
                    return obj
            except asyncio.TimeoutError as e:
                past_exceptions.append(
                    f'Timout a response for prompt {len(prompt)}\n{e}'
                )
                logging.warning(past_exceptions[-1])
                request_delay *= 2
            except aiohttp.client_exceptions.ClientOSError as e:
                past_exceptions.append(
                    f'Retrying a response for prompt {len(prompt)}\n{e}'
                )
                logging.warning(past_exceptions[-1])
                request_delay *= 2
            except json.decoder.JSONDecodeError as e:
                past_exceptions.append(
                    f'Got a bad response, {resp_text}. Retrying.\n{e}'
                )
                logging.debug(past_exceptions[-1])
                request_delay *= 2

            time.sleep(request_delay)

    @staticmethod
    async def async_request_many(
        server: str,
        api_key: str,
        prompts: List[str],
        timeout: Optional[int] = None,
        max_num_tries: int = -1,
        **kwargs,
    ):
        connector = aiohttp.TCPConnector(limit=0)
        timeout = aiohttp.ClientTimeout(total=timeout)
        async with aiohttp.ClientSession(
            timeout=timeout, connector=connector
        ) as session:
            tasks = []
            for prompt in prompts:
                tasks.append(
                    asyncio.ensure_future(
                        APIUtils.make_request(
                            session=session,
                            server=server,
                            api_key=api_key,
                            prompt=prompt,
                            num_retry_on_api_exception=max_num_tries,
                            **kwargs,
                        )
                    )
                )
            results = await asyncio.gather(*tasks)
            return results

    @staticmethod
    def request_many(
        server,
        api_key,
        prompts: List[str],
        timeout: Optional[int] = None,
        max_num_tries: int = -1,
        **kwargs,
    ) -> List[Dict[str, Any]]:
        return asyncio.run(APIUtils.async_request_many(server=server, api_key=api_key, prompts=prompts, timeout=timeout, max_num_tries=max_num_tries, **kwargs))  # type: ignore

    @staticmethod
    def compute_perplexities(
        observations: List[Message], results: List[Dict[str, Any]]
    ) -> Tuple[List[PPLMetric], List[PPLMetric]]:
        """
        Compute perplexities from API call.

        :param observations:
            incoming observations
        :param results:
            results from API call

        :return ppls:
            return list of perplexities
        """
        label_perplexities = []
        all_perplexities = []
        for obs, result in zip(observations, results):
            # need text offsets to figure out what comes from the prompt
            prompt_len = len(obs['prompt'])

            text_off = result['choices'][0]['logprobs']['text_offset']
            start_label = [i for i, off in enumerate(text_off) if off <= prompt_len]
            assert len(start_label) > 0
            start_label = start_label[-1]
            all_log_probs = [
                lp
                for lp in result['choices'][0]['logprobs']['token_logprobs']
                if lp is not None
            ]
            if not all(l <= 0 for l in all_log_probs):
                logging.warning(
                    f'Out of {len(all_log_probs)} log probs, {len([l for l in all_log_probs if l > 0])} are > 0. '
                    'Clamping to 0'
                )
                all_log_probs = [min(l, 0) for l in all_log_probs]

            log_probs = all_log_probs[start_label:]
            loss = -sum(log_probs)
            label_perplexities.append(PPLMetric(loss, len(log_probs)))
            all_perplexities.append(PPLMetric(-sum(all_log_probs), len(all_log_probs)))
        return label_perplexities, all_perplexities


class DisplayUtils:
    @staticmethod
    def display_observations(observations: Dict[Module, Dict[str, Any]]):
        """
        Print the observations nicely.
        """
        for module, obs in observations.items():
            if module == 'raw':
                continue
            print(
                f"{module.message_name()}\n{'-' * 80}\n{obs.get('prompt')}\n{'-' * 80}\n{'-' * 80}"
            )

    @staticmethod
    def display_act(act: Message):
        """
        Print the observations nicely.
        """
        print("=" * 80)
        print(f"{'-' * 20} DISPLAYING ACT FIELDS {'-' * 20}")
        print("=" * 80)
        for key, val in act.items():
            if not val:
                continue
            print(f"{key}\n{'-' * 80}\n{val}\n{'-' * 80}\n{'-' * 80}")



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from parlai.core.teachers import Teacher
from .build import build
from parlai.utils.io import PathManager
import json
import os
import random
import copy

WELCOME_MESSAGE = "Negotiate with your opponent to decide who gets how many items of each kind. There are three kinds of packages: Food, Water, and Firewood. Each has a quantity of 3. Try hard to get as much value as you can, while still leaving your partner satisfied and with a positive perception about you. If you fail to come to an agreement, both parties get 5 points. Refer to the following preference order and arguments for your negotiation: \n\nFood\nValue: {food_val} points for each package\nArgument: {food_argument}\n\nWater\nValue: {water_val} points for each package\nArgument: {water_argument}\n\nFirewood\nValue: {firewood_val} points for each package\nArgument: {firewood_argument}\n"


def get_welcome_values(part_info):

    value2points = {'High': 5, 'Medium': 4, 'Low': 3}

    issue2points = {v: value2points[k] for k, v in part_info['value2issue'].items()}
    issue2reason = {
        v: part_info['value2reason'][k] for k, v in part_info['value2issue'].items()
    }

    welcome_values = {}
    for issue in ['Food', 'Water', 'Firewood']:
        welcome_values[issue.lower() + '_val'] = issue2points[issue]
        welcome_values[issue.lower() + '_argument'] = issue2reason[issue]

    return welcome_values


def get_utterance_text(utterance):

    if utterance['text'] == '<DUMMY>':
        return ''

    # the utterance is not a dummy one at this point
    if utterance['text'] != 'Submit-Deal':
        # simply return it
        return utterance['text']

    # if it is a Submit-Deal -> attach task_data
    txt = f"{utterance['text']} What I get- Food:{utterance['task_data']['issue2youget']['Food']}, Water: {utterance['task_data']['issue2youget']['Water']}, Firewood: {utterance['task_data']['issue2youget']['Firewood']}; What you get- Food:{utterance['task_data']['issue2theyget']['Food']}, Water: {utterance['task_data']['issue2theyget']['Water']}, Firewood: {utterance['task_data']['issue2theyget']['Firewood']}"

    return txt


class CasinoTeacher(Teacher):
    """
    A negotiation teacher that loads the CaSiNo data from
    https://github.com/kushalchawla/CaSiNo.

    Each dialogue is converted into two datapoints, one from the perspective of each
    participant.
    """

    def __init__(self, opt, shared=None):
        super().__init__(opt, shared)
        self.datatype = opt['datatype'].split(':')[0]
        self.datatype_ = opt['datatype']
        self.random = self.datatype_ == 'train'
        build(opt)

        filename = self.datatype
        data_path = os.path.join(
            opt['datapath'], 'casino', 'casino_' + filename + '.json'
        )

        if shared and 'data' in shared:
            self.episodes = shared['episodes']
        else:
            self._setup_data(data_path)
        print(f"Total episodes: {self.num_episodes()}")

        # for ordered data in batch mode (especially, for validation and
        # testing), each teacher in the batch gets a start index and a step
        # size so they all process disparate sets of the data
        self.step_size = opt.get('batchsize', 1)
        self.data_offset = opt.get('batchindex', 0)

        self.reset()

    def _setup_data(self, data_path):
        print('loading: ' + data_path)
        with PathManager.open(data_path) as data_file:
            dialogues = json.load(data_file)

        episodes = []
        for dialogue in dialogues:

            # divide the dialogue into two perspectives, one for each participant
            episode = copy.deepcopy(dialogue)
            episode[
                'perspective'
            ] = 'mturk_agent_1'  # id of the agent whose perspective will be used in this dialog
            episodes.append(episode)

            episode = copy.deepcopy(dialogue)
            episode[
                'perspective'
            ] = 'mturk_agent_2'  # id of the agent whose perspective will be used in this dialog
            episodes.append(episode)

        self.episodes = episodes

        # add dummy data to ensure that every chat begins with a teacher utterance (THEM) and ends at the agent's utterance (YOU). This is done for uniformity while parsing the data. It makes the code simpler and easier to read than DealNoDeal counterpart.
        for ix, episode in enumerate(self.episodes):

            chat_logs = episode['chat_logs']
            perspective = episode['perspective']

            if chat_logs[0]['id'] == perspective:
                # chat must start with a teacher; add dummy utterance
                dummy_utterance = {
                    'text': '<DUMMY>',
                    'task_data': {},
                    'id': 'mturk_agent_1'
                    if perspective == 'mturk_agent_2'
                    else 'mturk_agent_2',
                }

                chat_logs = [dummy_utterance] + chat_logs

            if chat_logs[-1]['id'] != perspective:
                # chat must end with the agent; add dummy utterance
                dummy_utterance = {
                    'text': '<DUMMY>',
                    'task_data': {},
                    'id': 'mturk_agent_1'
                    if perspective == 'mturk_agent_1'
                    else 'mturk_agent_2',
                }

                chat_logs = chat_logs + [dummy_utterance]

            self.episodes[ix]['chat_logs'] = chat_logs

    def reset(self):
        super().reset()
        self.episode_idx = self.data_offset - self.step_size
        self.dialogue_idx = None
        self.perspective = None
        self.dialogue = None
        self.output = None
        self.expected_response = None
        self.epochDone = False

    def num_examples(self):
        """
        Lets return the the number of responses that an agent would generate in one
        epoch + 1 count for every output.

        This will include special utterances for submit-deal, accept-deal, and reject-
        deal.
        """
        num_exs = 0

        for episode in self.episodes:

            for utt in episode['chat_logs']:
                if utt['text'] != '<DUMMY>':
                    # skip the dummy utterances
                    num_exs += 1

        return (num_exs // 2) + len(
            self.episodes
        )  # since each dialogue was converted into 2 perspectives, one for each participant: see _setup_data

    def num_episodes(self):
        return len(self.episodes)

    def share(self):
        shared = super().share()
        shared['episodes'] = self.episodes
        return shared

    def observe(self, observation):
        """
        Process observation for metrics.
        """
        if self.expected_response is not None:
            self.metrics.evaluate_response(observation, self.expected_response)
            self.expected_response = None
        return observation

    def act(self):
        if self.dialogue_idx is not None:
            # continue existing conversation
            return self._continue_dialogue()
        elif self.random:
            # if random, then select the next random example
            self.episode_idx = random.randrange(len(self.episodes))
            return self._start_dialogue()
        elif self.episode_idx + self.step_size >= len(self.episodes):
            # end of examples
            self.epochDone = True
            return {'episode_done': True}
        else:
            # get next non-random example
            self.episode_idx = (self.episode_idx + self.step_size) % len(self.episodes)
            return self._start_dialogue()

    def _start_dialogue(self):
        """
        Starting a dialogue should be the same as continuing a dialogue but with just
        one difference: it will attach the welcome note to the teacher's utterance.

        Each dialogue has two agents possible: mturk_agent_1 or mturk_agent_2. One of
        them will act as the perspective for this episode.
        """

        episode = self.episodes[self.episode_idx]
        self.perspective = episode['perspective']
        self.other_id = (
            'mturk_agent_1' if self.perspective == 'mturk_agent_2' else 'mturk_agent_2'
        )

        part_info = episode['participant_info'][self.perspective]
        part_info_other = episode['participant_info'][self.other_id]

        welcome_values = get_welcome_values(part_info)
        welcome = WELCOME_MESSAGE.format(
            food_val=welcome_values['food_val'],
            water_val=welcome_values['water_val'],
            firewood_val=welcome_values['firewood_val'],
            food_argument=welcome_values['food_argument'],
            water_argument=welcome_values['water_argument'],
            firewood_argument=welcome_values['firewood_argument'],
        )

        self.dialogue = episode['chat_logs']
        self.output = {
            'your_points_scored': part_info['outcomes']['points_scored'],
            'how_satisfied_is_your_partner': part_info_other['outcomes'][
                'satisfaction'
            ],
            'how_much_does_your_partner_like_you': part_info_other['outcomes'][
                'opponent_likeness'
            ],
        }

        self.dialogue_idx = -1

        action = self._continue_dialogue()
        if action['text']:
            # This is non-empty; meaning the teacher starts the conversation and has something to say.
            action['text'] = f"{welcome}\n{action['text']}"
        else:
            # text is empty, meaning that the teacher did not start the conversation but the empty string is just a result of the dummy teacher utterance added in _setup_data
            action['text'] = welcome

        action['meta-info'] = welcome_values

        return action

    def _continue_dialogue(self):
        """
        Return an action object.

        From the perspective of a specific agent's id, all utterances authored by the
        other agent are coming from the teacher as the text of the action object, and
        all utterances authored by this agent appear as the labels.
        """
        action = {}
        # Fill in teacher's message (THEM)
        self.dialogue_idx += 1
        if self.dialogue_idx < len(self.dialogue):
            # this is a usual dialogue teacher-agent pair; return the teacher's utterance as action text.
            utterance = self.dialogue[self.dialogue_idx]
            assert utterance['id'] != self.perspective
            utterance_text = get_utterance_text(
                utterance
            )  # will take care of special submit-deal utterance and dummy utterances
            action['text'] = utterance_text

            if action['text'] == 'Reject-Deal':
                # merge with the next dialogue_idx since that is from the same participant while this code assumes alternative utterances.
                self.dialogue_idx += 1  # we know that this will be valid
                utterance = self.dialogue[self.dialogue_idx]
                assert utterance['id'] != self.perspective
                utterance_text = get_utterance_text(
                    utterance
                )  # will take care of special submit-deal utterance and dummy utterances
                action['text'] = action['text'] + ' ' + utterance_text
        else:
            # the primary dialogue is over; now is the time to return the output of this dialogue
            action[
                'text'
            ] = f"Your points scored: {self.output['your_points_scored']}, How satisfied is your partner: {self.output['how_satisfied_is_your_partner']}, How much does your partner like you: {self.output['how_much_does_your_partner_like_you']}"

        # Fill in learner's response (YOU)
        self.dialogue_idx += 1
        self.expected_response = None
        if self.dialogue_idx < len(self.dialogue):
            # usual dialogue going on; return the agent's utterance as the labels
            utterance = self.dialogue[self.dialogue_idx]
            assert (
                utterance['id'] == self.perspective
            ), f"id: {utterance['id']}, perspect: {self.perspective}"
            utterance_text1 = get_utterance_text(
                utterance
            )  # will take care of special submit-deal utterance and dummy utterances

            utterance_text2 = ''
            if utterance_text1 == 'Reject-Deal':
                # merge with the next dialogue_idx since that is from the same participant while this code assumes alternative utterances.
                self.dialogue_idx += 1  # we know that this will be valid
                utterance = self.dialogue[self.dialogue_idx]
                assert utterance['id'] == self.perspective
                utterance_text2 = get_utterance_text(
                    utterance
                )  # will take care of special submit-deal utterance and dummy utterances

            self.expected_response = (
                [utterance_text1 + ' ' + utterance_text2]
                if (utterance_text1 + ' ' + utterance_text2).strip()
                else None
            )
        else:
            # no label required when the primary dialogue is complete
            pass

        if self.expected_response:
            # since labels is automatically renamed to eval_labels for valid/test, doing just this takes care of everything. Ensures that labels can atleast be accessed regardless of the datatype.
            action['labels'] = self.expected_response

        if self.dialogue_idx >= len(self.dialogue):
            self.dialogue_idx = None
            action['episode_done'] = True
        else:
            action['episode_done'] = False

        return action


class DefaultTeacher(CasinoTeacher):
    pass



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
# Download and build the data if it does not exist.

import parlai.core.build_data as build_data
import os
from parlai.core.build_data import DownloadableFile

RESOURCES = [
    DownloadableFile(
        'https://github.com/wise-east/spolin/raw/master/data/spolin-train-acl.json',
        'spolin-train-acl.json',
        'db3e1470f28c9dbf8423c2a8f5e4a52aad932f68a6f86d376597f0f36b845439',
        zipped=False,
    ),
    DownloadableFile(
        'https://github.com/wise-east/spolin/raw/master/data/spolin-train.json',
        'spolin-train.json',
        '3b87ce000d0900a1811298f04ebff1922368f99b68eb9b1d9f436aaf53027f64',
        zipped=False,
    ),
    DownloadableFile(
        'https://github.com/wise-east/spolin/raw/master/data/spolin-valid.json',
        'spolin-valid.json',
        '0db6afb5ddbcbfff974f5ca5f2b575433cc02d3a65662f808e12bd82b634e141',
        zipped=False,
    ),
]


def build(opt):
    dpath = os.path.join(opt['datapath'], 'spolin')
    version = None

    if not build_data.built(dpath, version_string=version):
        print('[building data: ' + dpath + ']')
        if build_data.built(dpath):
            # An older version exists, so remove these outdated files.
            build_data.remove_dir(dpath)
        build_data.make_dir(dpath)

        # Download the data.
        for downloadable_file in RESOURCES:
            downloadable_file.download_file(dpath)

        # Mark the data as built.
        build_data.mark_done(dpath, version_string=version)



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import json
import os
from typing import Optional

from parlai.core.opt import Opt
from parlai.core.params import ParlaiParser
from parlai.core.teachers import DialogTeacher

from .build import build


class ProsocialDialogSafetyTeacher(DialogTeacher):
    """
    Safety Teacher for ProsocialDialog Data https://github.com/skywalker023/prosocial-dialog
    set --one-turn to true for just one turn without the context.
    """

    @classmethod
    def add_cmdline_args(
        cls, parser: ParlaiParser, partial_opt: Optional[Opt] = None
    ) -> ParlaiParser:
        super().add_cmdline_args(parser, partial_opt)
        agent = parser.add_argument_group('prosocial dialog safety options')
        agent.add_argument(
            '--one-turn',
            type=bool,
            default=False,
            help="Whether or not to have the text include the context if it exists or just single turn",
        )
        return parser

    def __init__(self, opt, shared=None):
        self.datatype = opt['datatype']
        build(opt)
        self.datatype = opt['datatype'].split(':')[0]
        opt['datafile'] = os.path.join(
            opt['datapath'], 'prosocial_dialog', self.datatype + '.json'
        )
        self.id = 'prosocial_dialog'
        super().__init__(opt, shared)

    def setup_data(self, path):
        print('loading: ' + path)
        f = open(path)
        self.json_data = json.load(f)
        f.close()

        for exs in self.json_data:
            texts = []
            for ex in exs:
                texts.append(ex['text'])
                if self.opt['one_turn']:
                    x = ex['text']
                else:
                    x = '\n'.join(texts)
                texts.append(ex['labels'][0])
                y = ex['safety_label']
                m = {'text': x, 'labels': y}
                yield m, True

    def num_episodes(self):
        return sum([len(x) for x in self.json_data])

    def num_examples(self):
        return sum([len(x) for x in self.json_data])


class ProsocialDialogBinarySafetyTeacher(ProsocialDialogSafetyTeacher):
    """
    Binary Safety Teacher for ProsocialDialog Data https://github.com/skywalker023/prosocial-dialog
    Casual is __ok__ and Needs Caution and Needs Intervention is __notok__
    """

    def setup_data(self, path):
        print('loading: ' + path)
        f = open(path)
        self.json_data = json.load(f)
        f.close()

        for exs in self.json_data:
            texts = []
            for ex in exs:
                texts.append(ex['text'])
                if self.opt['one_turn']:
                    x = ex['text']
                else:
                    x = '\n'.join(texts)
                texts.append(ex['labels'][0])
                y = "__ok__" if ex['safety_label'] == "__casual__" else "__notok__"
                m = {'text': x, 'labels': y}
                yield m, True


class ProsocialDialogTeacher(DialogTeacher):
    """
    Teacher for ProsocialDialog Data https://github.com/skywalker023/prosocial-dialog
    """

    @classmethod
    def add_cmdline_args(
        cls, parser: ParlaiParser, partial_opt: Optional[Opt] = None
    ) -> ParlaiParser:
        super().add_cmdline_args(parser, partial_opt)
        return parser

    def __init__(self, opt, shared=None):
        self.datatype = opt['datatype']
        build(opt)
        self.datatype = opt['datatype'].split(':')[0]
        opt['datafile'] = os.path.join(
            opt['datapath'], 'prosocial_dialog', self.datatype + '.json'
        )
        self.id = 'prosocial_dialog'
        super().__init__(opt, shared)

    def setup_data(self, path):
        print('loading: ' + path)
        f = open(path)
        self.json_data = json.load(f)
        f.close()

        for exs in self.json_data:
            for ex in exs:
                x = ex['text']
                y = ex['labels']
                m = {'text': x, 'labels': y}
                yield m, ex['episode_done']

    def num_episodes(self):
        return len(self.json_data)

    def num_examples(self):
        return sum([len(x) for x in self.json_data])


class DefaultTeacher(ProsocialDialogTeacher):
    pass



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
"""
Utility Functions.
"""
from collections import Counter
from enum import Enum
import torch
from typing import Callable, List, Tuple

from parlai.core.torch_agent import Batch
from parlai.tasks.wizard_of_wikipedia.agents import TOKEN_KNOWLEDGE, TOKEN_END_KNOWLEDGE
import parlai.utils.logging as logging

try:
    from nltk.corpus import stopwords

    STOP_WORDS = stopwords.words('english')
    import nltk
except (ModuleNotFoundError, LookupError):
    logging.error("Please install NLTK: pip install nltk")
    logging.error("Then run `nltk.download('stopwords')`")
    nltk = None
    stopwords = None
    STOP_WORDS = []

try:
    import spacy
except ModuleNotFoundError:
    logging.error("Please install spacy: pip install spacy")
    spacy = None

##############
# Zoo Models #
##############
R2C2_BASE_400M = 'zoo:seeker/r2c2_base_400M/model'
R2C2_BASE_3B = 'zoo:seeker/r2c2_base_3B/model'
R2C2_BLENDERBOT_400M = 'zoo:seeker/r2c2_blenderbot_400M/model'
R2C2_BLENDERBOT_3B = 'zoo:seeker/r2c2_blenderbot_3B/model'
SEEKER_DIALOGUE_400M = 'zoo:seeker/seeker_dialogue_400M/model'
SEEKER_DIALOGUE_3B = 'zoo:seeker/seeker_dialogue_3B/model'
SEEKER_LM_DIALOGUE_3B = 'zoo:seeker/seeker_lm_dialogue_3B/model'
SEEKER_LM_MED = 'zoo:seeker/seeker_lm_med/model'
SEEKER_LM_LARGE = 'zoo:seeker/seeker_lm_large/model'
SEEKER_LM_XL = 'zoo:seeker/seeker_lm_xl/model'

##################
# Control Tokens #
##################
GENERATE_QUERY = '__generate-query__'
IS_SEARCH_REQUIRED = '__is-search-required__'
DO_SEARCH = '__do-search__'
DO_NOT_SEARCH = '__do-not-search__'

nlp = None


class SearchDecision(Enum):
    """
    Abstracting how to decide to search.
    """

    ALWAYS = 'always'
    NEVER = 'never'
    COMPUTE = 'compute'


def drm_get_batch_context(self, batch: Batch, orig_fun: Callable):
    """
    Override the get context method for the DRM.

    This override removes the knowledge response from the set of tokens being blocked in
    the dialogue response.
    """
    ctxts = orig_fun(batch)
    knowledge_start_id = self.dict.txt2vec(TOKEN_KNOWLEDGE)
    knowledge_end_id = self.dict.txt2vec(TOKEN_END_KNOWLEDGE)

    def mask_ctxttensor_between_sublists(
        ctxts: torch.Tensor, sub1: List[int], sub2: List[int]
    ) -> torch.Tensor:
        """
        Generate a mask that masks out the context between sub1 and sub2.
        """
        mask = []
        for ctxt in ctxts:
            mask_idxs = []
            should_copy = False
            idx_pointer = 0
            id_to_match = sub1
            for j, token in enumerate(ctxt.cpu().numpy()):
                if token == id_to_match[idx_pointer]:
                    idx_pointer += 1
                    if idx_pointer == 1 and id_to_match == sub1:
                        mask_idxs.append([j])
                    elif idx_pointer >= len(id_to_match):
                        should_copy = id_to_match == sub1
                        idx_pointer = 0
                        id_to_match = sub2 if (id_to_match == sub1) else sub1
                        mask_idxs[-1].append(j)
                    else:
                        mask_idxs[-1].append(j)
                elif should_copy:
                    assert isinstance(mask_idxs[-1], list)
                    mask_idxs[-1].append(j)
                elif idx_pointer > 0:
                    idx_pointer = 0
                    del mask_idxs[-1]
            mask.append(
                [
                    0 if idx in [i for sl in mask_idxs for i in sl] else 1
                    for idx in range(len(ctxt))
                ]
            )
        return torch.LongTensor(mask).to(ctxts.device)

    ctxts *= mask_ctxttensor_between_sublists(
        ctxts, knowledge_start_id, knowledge_end_id
    )
    return ctxts


def krm_get_batch_context(
    self, batch: Batch, orig_fun: Callable[[Batch], torch.Tensor]
):
    """
    Monkey-patch get_batch_context for the KRM to block on the prior knowledge
    responses.
    """
    ctxts = orig_fun(batch)
    if 'prior_knowledge_responses_vec' in batch:
        ctxts = torch.cat([ctxts, batch.prior_knowledge_responses_vec.to(ctxts)], dim=1)
    return ctxts


def krm_get_batch_context_only_knowledge(
    self, batch: Batch, orig_fun: Callable[[Batch], torch.Tensor]
):
    """
    Monkey-patch get_batch_context for the KRM to block on the prior knowledge
    responses.

    Discard conversational context.
    """
    ctxts = orig_fun(batch)
    if 'prior_knowledge_responses_vec' in batch:
        ctxts = batch.prior_knowledge_responses_vec.to(ctxts)
    return ctxts


def extract_entities(
    sentence: str,
    pos: Tuple[str] = ('PROPN', 'NOUN'),
    use_named_entities: bool = True,
    use_noun_chunks: bool = True,
) -> List[str]:
    """
    Given a sentence, extract the entities from the sentence.

    :param sentence:
        provided sentence
    :param pos:
        parts of speech to look at
    :param use_named_entities:
        whether to include named entities
    :param use_noun_chunks:
        whether to include noun chunks.

    :return entities:
        return list of entities.
    """
    global nlp
    if nlp is None:
        logging.info('Loading spacy once')
        try:
            assert spacy is not None
            nlp = spacy.load("en_core_web_sm")
        except Exception:
            raise RuntimeError(
                'Please download: python -m spacy download en_core_web_sm'
            )
    doc = nlp(sentence)
    results = []
    if pos:
        for token in doc:
            if token.pos_ in pos:
                results.append(token)
    if use_named_entities:
        for ent in doc.ents:
            results.append(ent)
    if use_noun_chunks:
        for chunk in doc.noun_chunks:
            if chunk.text.lower() not in STOP_WORDS:
                results.append(chunk)
    results = list(set([r.text for r in results]))
    return results


def calc_f1_msmarco(pred: str, gold_items: List[str]) -> float:
    """
    Calculate F1 overlap between prediction sentence and gold labels.

    :param pred:
        prediction string
    :param gold_items:
        list of gold items

    :return f1_overlap:
    """
    try:
        assert nltk is not None
        pred_items = nltk.word_tokenize(pred)
    except IndexError:
        # malformed prediction; return 0
        return 0
    common = Counter(gold_items) & Counter(pred_items)
    num_same = sum(common.values())
    if num_same == 0:
        return 0
    precision = 1.0 * num_same / len(pred_items)
    recall = 1.0 * num_same / len(gold_items)
    f1 = (2 * precision * recall) / (precision + recall)
    return f1


def calc_f1_msc(pred, gold_items):
    """
    Calculate F1 overlap between prediction sentence and gold labels.

    :param pred:
        prediction string
    :param gold_items:
        list of gold items

    :return f1_overlap:
    """
    try:
        assert nltk is not None
        pred_items = nltk.word_tokenize(pred)
    except IndexError:
        # malformed prediction; return 0
        return 0
    common = Counter(gold_items) & Counter(pred_items)
    num_same = sum(common.values())
    if num_same == 0:
        return 0
    precision = 1.0 * num_same / len(pred_items)
    recall = 1.0 * num_same / len(gold_items)
    f1 = (2 * precision * recall) / (precision + recall)

    context_ent = extract_entities(' '.join(gold_items))
    label_ent = extract_entities(pred)
    ents = set(context_ent).intersection(label_ent)
    if len(ents) == 0:
        f1 = 0

    return f1


def remove_possible_title_from_text(
    text: str, title: str, min_title_length: int = 3, overlap_ratio: float = 0.5
) -> str:
    """
    Remove title from text document.

    :param text:
        text string
    :param title:
        title to remove
    :param min_title_length:
        minimum length of title to remove
    :param overlap_ratio:
        minimum ratio required to remove.

    :return cleaned:
        return cleaned text
    """

    def _high_intersection(s1, s2):
        return len(s1.intersection(s2)) > overlap_ratio * len(s2)

    titile_tokesn = set(title.lower().split(' '))
    if (len(titile_tokesn) < min_title_length) or (len(text) <= len(title)):
        return text
    text_beginning_tokens = set(text.lower().split(' '))
    if _high_intersection(text_beginning_tokens, titile_tokesn):
        text = text[len(title) :]
    return text



#!/usr/bin/env python3
# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
from typing import List
from parlai.core.opt import Opt
import parlai.utils.logging as logging

from projects.bb3.tasks.r2c2_knowledge_tasks import (
    MsMarcoSearchKnowledgeTeacher as msmarco_skm,
    NQSearchKnowledgeTeacher as nq_skm,
    NQOpenSearchKnowledgeTeacher as nqopen_skm,
    NQOpenDialoguesSearchKnowledgeTeacher as nqopendialogues_skm,
    TriviaQASearchKnowledgeTeacher as triviaqa_skm,
    SquadSearchKnowledgeTeacher as squad_skm,
    WowSearchKnowledgeTeacher as wow_skm,
    WoiSearchKnowledgeTeacher as woi_skm,
    Convai2EntityKnowledgeTeacher as convai2_ckm,
    EDEntityKnowledgeTeacher as ed_ckm,
    MSCEntityKnowledgeTeacher as msc_ckm,
    BSTEntityKnowledgeTeacher as bst_ckm,
    Convai2MemoryKnowledgePersOverlapTeacher as convai2_pers_mkm,
    EDMemoryKnowledgePersOverlapTeacher as ed_pers_mkm,
    BSTMemoryKnowledgePersOverlapTeacher as bst_pers_mkm,
    MSCMemoryKnowledgePersOverlapTeacher as msc_pers_mkm,
    Convai2MemoryKnowledgeUttOverlapTeacher as convai2_utt_mkm,
    EDMemoryKnowledgeUttOverlapTeacher as ed_utt_mkm,
    BSTMemoryKnowledgeUttOverlapTeacher as bst_utt_mkm,
    MSCMemoryKnowledgeUttOverlapTeacher as msc_utt_mkm,
)

from projects.bb3.tasks.module_level_tasks import BB3TeacherMixin

from parlai.tasks.fits.agents import FitsBaseTeacher as FitsKnowledgeTeacher


def get_knowledge_opt_mutators(opt: Opt):
    mutators = '+'.join(
        ['prefix_speakers_opt', 'format_knowledge_tasks_for_decoder_only']
    )
    if opt.get('mutators'):
        mutators = '+'.join([opt['mutators'], mutators])
    logging.warning(f'overriding mutators to {mutators}')
    return mutators


class Convai2EntityKnowledgeTeacher(convai2_ckm):
    def get_special_mutators(self, opt) -> str:
        return '+'.join(
            [
                super().get_special_mutators(opt),
                'prefix_speakers_opt',
                'format_knowledge_tasks_for_decoder_only',
            ]
        )


class EDEntityKnowledgeTeacher(ed_ckm):
    def get_special_mutators(self, opt) -> str:
        return '+'.join(
            [
                super().get_special_mutators(opt),
                'prefix_speakers_opt',
                'format_knowledge_tasks_for_decoder_only',
            ]
        )


class MSCEntityKnowledgeTeacher(msc_ckm):
    def get_special_mutators(self, opt) -> str:
        return '+'.join(
            [
                super().get_special_mutators(opt),
                'prefix_speakers_opt',
                'format_knowledge_tasks_for_decoder_only',
            ]
        )


class BSTEntityKnowledgeTeacher(bst_ckm):
    def get_special_mutators(self, opt) -> str:
        return '+'.join(
            [
                super().get_special_mutators(opt),
                'prefix_speakers_opt',
                'format_knowledge_tasks_for_decoder_only',
            ]
        )


class Convai2MemoryKnowledgePersOverlapTeacher(convai2_pers_mkm):
    def __init__(self, opt: Opt, shared=None) -> None:
        opt['mutators'] = get_knowledge_opt_mutators(opt)
        super().__init__(opt, shared)


class EDMemoryKnowledgePersOverlapTeacher(ed_pers_mkm):
    def __init__(self, opt: Opt, shared=None) -> None:
        opt['mutators'] = get_knowledge_opt_mutators(opt)
        super().__init__(opt, shared)


class MSCMemoryKnowledgePersOverlapTeacher(msc_pers_mkm):
    def __init__(self, opt: Opt, shared=None) -> None:
        opt['mutators'] = get_knowledge_opt_mutators(opt)
        super().__init__(opt, shared)


class BSTMemoryKnowledgePersOverlapTeacher(bst_pers_mkm):
    def __init__(self, opt: Opt, shared=None) -> None:
        opt['mutators'] = get_knowledge_opt_mutators(opt)
        super().__init__(opt, shared)


class Convai2MemoryKnowledgeUttOverlapTeacher(convai2_utt_mkm):
    def __init__(self, opt: Opt, shared=None) -> None:
        opt['mutators'] = get_knowledge_opt_mutators(opt)
        super().__init__(opt, shared)


class EDMemoryKnowledgeUttOverlapTeacher(ed_utt_mkm):
    def __init__(self, opt: Opt, shared=None) -> None:
        opt['mutators'] = get_knowledge_opt_mutators(opt)
        super().__init__(opt, shared)


class MSCMemoryKnowledgeUttOverlapTeacher(msc_utt_mkm):
    def __init__(self, opt: Opt, shared=None) -> None:
        opt['mutators'] = get_knowledge_opt_mutators(opt)
        super().__init__(opt, shared)


class BSTMemoryKnowledgeUttOverlapTeacher(bst_utt_mkm):
    def __init__(self, opt: Opt, shared=None) -> None:
        opt['mutators'] = get_knowledge_opt_mutators(opt)
        super().__init__(opt, shared)


class MsMarcoSearchKnowledgeTeacher(msmarco_skm):
    def get_special_mutators(self) -> List[str]:
        return super().get_special_mutators() + [
            'prefix_speakers_opt',
            'format_knowledge_tasks_for_decoder_only_reduced_docs',
        ]


class NQSearchKnowledgeTeacher(nq_skm):
    def get_special_mutators(self) -> List[str]:
        return super().get_special_mutators() + [
            'prefix_speakers_opt',
            'format_knowledge_tasks_for_decoder_only_reduced_docs',
        ]


class NQOpenSearchKnowledgeTeacher(nqopen_skm):
    def get_special_mutators(self) -> List[str]:
        return super().get_special_mutators() + [
            'prefix_speakers_opt',
            'format_knowledge_tasks_for_decoder_only_reduced_docs',
        ]


class NQOpenDialoguesSearchKnowledgeTeacher(nqopendialogues_skm):
    def get_special_mutators(self) -> List[str]:
        return super().get_special_mutators() + [
            'prefix_speakers_opt',
            'format_knowledge_tasks_for_decoder_only_reduced_docs',
        ]


class TriviaQASearchKnowledgeTeacher(triviaqa_skm):
    def get_special_mutators(self) -> List[str]:
        return super().get_special_mutators() + [
            'prefix_speakers_opt',
            'format_knowledge_tasks_for_decoder_only_reduced_docs',
        ]


class SquadSearchKnowledgeTeacher(squad_skm):
    def get_special_mutators(self) -> List[str]:
        return super().get_special_mutators() + [
            'prefix_speakers_opt',
            'format_knowledge_tasks_for_decoder_only_reduced_docs',
        ]


class WowSearchKnowledgeTeacher(wow_skm):
    def get_special_mutators(self) -> List[str]:
        return super().get_special_mutators() + [
            'prefix_speakers_opt',
            'format_knowledge_tasks_for_decoder_only_reduced_docs',
        ]


class WoiSearchKnowledgeTeacher(woi_skm):
    def get_special_mutators(self) -> List[str]:
        return super().get_special_mutators() + [
            'prefix_speakers_opt',
            'format_knowledge_tasks_for_decoder_only_reduced_docs',
        ]


class FitsSearchKnowledgeTeacher(BB3TeacherMixin, FitsKnowledgeTeacher):
    def __init__(self, opt, shared=None):
        mutators = '+'.join(
            [
                'flatten_gold_knowledge_response_mutator',
                'fits_remove_special_toks_mutator',
                'add_selected_sentences_mutator',
                'prompt_knowledge_mutator',
                'prefix_speakers_opt',
                'format_knowledge_tasks_for_decoder_only_reduced_docs',
            ]
        )
        if opt.get('mutators'):
            mutators = '+'.join([mutators, opt['mutators']])
        logging.warning(f'overriding mutators to {mutators}')
        opt['mutators'] = mutators
        super().__init__(opt, shared)



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.


import parlai.core.build_data as build_data
import os
from parlai.core.build_data import DownloadableFile

URL = "https://raw.githubusercontent.com/chaochun/nlu-asdiv-dataset/master/dataset/ASDiv.xml"

RESOURCES = [
    # raw data file
    DownloadableFile(
        URL,
        "ASDiv.xml",
        "ef8904068482919ac48c8eeaaf6df344b8a308ba66d048c2d4d87eab82dc4929",
        zipped=False,
    ),
]

VERSION = "1.0"


def build(opt):
    # get path to data directory
    dpath = os.path.join(opt["datapath"], "asdiv")
    # define version if any
    version = VERSION

    # check if data had been previously built
    if not build_data.built(dpath, version_string=version):
        print("[building data: " + dpath + "]")

        # make a clean directory if needed
        if build_data.built(dpath):
            # an older version exists, so remove these outdated files.
            build_data.remove_dir(dpath)
        build_data.make_dir(dpath)

        # Download the data.
        for downloadable_file in RESOURCES:
            downloadable_file.download_file(dpath)

        # mark the data as built
        build_data.mark_done(dpath, version_string=version)



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from enum import Enum
import nltk
from nltk.corpus import stopwords


class GraphMutations(Enum):
    NO_MUTATION = 0
    DEL = 1
    ADD = 2


try:
    stopwords.words('english')
except LookupError:
    nltk.download('stopwords')


# The set of words in the description of graph that we skip while checking
# for overlap between knowledge graph vertices and the description.
# We remove the word "you" from this set becuase otherwise the edges that have you (player) vertices,
# which are quite important, will be reduced to empty vertices, and thus dropped.
GRAPH_VERT_SKIP_TOKEN = set(stopwords.words('english')) - {'you'}

LOCATION_NAME = 'loc-name'
LOCATION_DESCRIPTION = 'loc-desc'
SURROUNDING_OBJECTS = 'surr_obj'
KNOWLEDGE_GRAPH = 'KG'
ACTION = 'action'
OBSERVATION = 'obs'

# The delimiter characetrs between members of a set (eg objects, graph edges, etc.)
SET_MEMBERS_DELIM = ' ; '
GRAPH_DELIM = ','

EMPTY_GRAPH_TOKEN = '__empty-graph__'



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
"""
Quick script for dumping out relevant conversation ids from GoogleSGD.
"""

from parlai.core.params import ParlaiParser
from parlai.core.script import ParlaiScript, register_script
from parlai.tasks.google_sgd_simulation_splits.agents import GoogleSgdOutDomainParser
from parlai.core.tod.tod_agents import TodStructuredDataParser

import parlai
import os
import json
import random

PARLAI_DATA_PATH = os.path.dirname(os.path.dirname(parlai.__file__)) + "/data"


class GrabEpisodes(GoogleSgdOutDomainParser, TodStructuredDataParser):
    def get_agent_type_suffix(self):
        return "GrabEpisodes"


def setup_args(parser=None):
    if not parser:
        parser = ParlaiParser(False, False)
    group = parser.add_argument_group("Get active learning samples script")
    group.add_argument(
        "--find-random-al-samples",
        type=bool,
        default=False,
        help="Get active learning samples randomly or few shot",
    )
    group.add_argument(
        "--input-processed-stats",
        type=str,
        help="Processed stats file from the `get_passing_only` script",
    )
    group.add_argument(
        "--processed-stats-section",
        type=str,
        default="MISSES",
        help="Which section from the `get_passing_only` script will we use to rank. (MISSES and FRACTIONAL current options)",
    )
    group.add_argument(
        "--num-apis-to-get",
        default=8,
        type=int,
        help="Number of api descriptions we want to find",
    )
    group.add_argument(
        "--existing-al-files",
        nargs="*",
        type=str,
        help="Existing active learning files (ie, for running multiple iterations of learning)",
    )
    group.add_argument(
        "--cumulative-al",
        type=bool,
        default=True,
        help="Uses active learning files from '--existing-al-files' cumulatively (as in, will append all prior dialog ids for next, rather than excluding)",
    )
    group.add_argument(
        "--al-output-file",
        default=None,
        help="Output file. Will put into 'results' in active run directory otherwise.",
    )
    return parser


@register_script("get_al_samples_for_gsgd_script")
class GetAlSamplesForGsgdScript(ParlaiScript):
    @classmethod
    def setup_args(cls):
        return setup_args()

    def run(self):
        existing_al_ids = self.get_existing_al_ids()
        # NOTE: The inidivdual get_al_samples funcitons are responsible for dealing with cumulative
        if self.opt["find_random_al_samples"]:
            save_me = self.get_al_samples_random(existing_al_ids)
        else:
            save_me = self.get_al_samples_from_processed(existing_al_ids)
        out_file = self.opt.get("al_output_file")
        if not out_file:
            out_file = "result"
        with open(out_file, "w+") as f:
            json.dump(save_me, f, indent=4)
        print("Saved AL samples to ", out_file)

    def get_al_samples_random(self, existing_al_ids):
        save_me = {}
        for datatype in ["train", "valid", "test"]:
            unfiltered_episodes = self.get_gsgd_episodes_for_datatype(datatype)
            filtered_episodes = [
                x
                for x in unfiltered_episodes
                if x.extras["dialogue_id"] not in existing_al_ids
            ]
            samples = random.Random(42).sample(
                filtered_episodes, self.opt["num_apis_to_get"]
            )
            if self.opt["cumulative_al"]:
                old = [
                    x
                    for x in unfiltered_episodes
                    if x.extras["dialogue_id"] in existing_al_ids
                ]
                samples.extend(old)
            save_me[datatype] = {
                episode.extras["dialogue_id"]: episode.goal_calls_utt
                for episode in samples
            }
        return save_me

    def get_al_samples_from_processed(self, existing_al_ids):
        wanted_apis = self.get_wanted_apis()
        save_me = {}
        for datatype in ["train", "valid", "test"]:
            found = [False] * len(wanted_apis)
            save_me[datatype] = {}
            for episode in self.get_gsgd_episodes_for_datatype(datatype):
                if episode.extras["dialogue_id"] in existing_al_ids.get(datatype, []):
                    if self.opt["cumulative_al"]:
                        save_me[datatype][
                            episode.extras["dialogue_id"]
                        ] = episode.goal_calls_utt
                    continue
                here = False
                for idx, val in enumerate(found):
                    if not val and not here:
                        here = True
                        for api_name_slots in wanted_apis[idx]:
                            api_name = api_name_slots[0]
                            slots = api_name_slots[1]
                            if api_name in episode.goal_calls_utt:
                                for slot in slots:
                                    if slot not in episode.goal_calls_utt:
                                        here = False
                            else:
                                here = False
                        if here:
                            found[idx] = True
                            save_me[datatype][
                                episode.extras["dialogue_id"]
                            ] = episode.goal_calls_utt
        return save_me

    def get_gsgd_episodes_for_datatype(self, datatype):
        datapath = PARLAI_DATA_PATH
        if "datapath" in self.opt:
            datapath = self.opt["datapath"]
        elif "parlai_home" in self.opt:
            datapath = self.opt["parlai_home"] + "/data"
        opt = {
            "datatype": datatype,
            "datapath": datapath,
            "gsgd_domains": "all",
            "n_shot": -1,
            "episodes_randomization_seed": -1,
        }
        return GrabEpisodes(opt).episodes

    def get_existing_al_ids(self):
        existing_al_files = self.opt.get("existing_al_files")
        if existing_al_files is None:
            return {}
        existing_al_ids = {}
        for existing in existing_al_files:
            with open(existing) as f:
                data = json.load(f)
            for datatype, blob in data.items():
                if datatype not in existing_al_ids:
                    existing_al_ids[datatype] = []
                for dialog_id in blob:
                    existing_al_ids[datatype].append(dialog_id)
        return existing_al_ids

    def get_wanted_apis(self):
        unprocessed_lines = []
        with open(self.opt["input_processed_stats"]) as f:
            lines = f.readlines()
            idx = 0
            while self.opt["processed_stats_section"] not in lines[idx]:
                idx += 1
            unprocessed_lines = lines[idx + 1 : idx + 1 + self.opt["num_apis_to_get"]]

        processed = []
        for row_raw in unprocessed_lines:
            print(row_raw)
            row = json.loads(json.loads(row_raw.strip())[0])
            slots = [x for x in row if "api_name" not in x]
            for x in row:
                if "api_name" in x:
                    processed.append((x.replace("api_name:", ""), slots))
        return processed


if __name__ == "__main__":
    GetAlSamplesForGsgdScript.main()



#!/usr/bin/env python3
# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
from parlai.core.opt import Opt
import parlai.utils.logging as logging
import parlai.tasks.convai2.agents as convai2
import parlai.tasks.blended_skill_talk.agents as bst
import parlai.tasks.empathetic_dialogues.agents as ed
import parlai.tasks.msc.agents as msc
import parlai.tasks.funpedia.agents as funpedia
import parlai.tasks.google_sgd.agents as google_sgd
import parlai.tasks.taskmaster.agents as taskmaster
import parlai.tasks.taskmaster2.agents as taskmaster2
import parlai.tasks.taskmaster3.agents as taskmaster3
import parlai.tasks.saferdialogues.agents as Saferdialogues
import parlai.tasks.wizard_of_wikipedia.agents as wow
import parlai.tasks.wizard_of_internet.agents as woi
import parlai.tasks.light_dialog.agents as light
import parlai.tasks.light_dialog_wild.agents as light_wild
import parlai.tasks.style_gen.agents as style_gen


from projects.seeker.tasks.dialogue import (
    WoiDialogueTeacher as woi_srm,
    WowDialogueTeacher as wow_srm,
    MsMarcoDialogueTeacher as msmarco_srm,
)

from parlai.tasks.fits.agents import FitsBaseTeacher as FitsDialogueTeacher
import parlai.tasks.fits.mutators  # type: ignore

from projects.bb3.tasks.module_level_tasks import BB3TeacherMixin
from projects.bb3.tasks.r2c2_knowledge_tasks import (
    get_memory_knowledge_utt_overlap_task_mutator,
    get_memory_knowledge_pers_overlap_task_mutator,
)

##################
# SearchDialogue #
##################


class WowSearchDialogueTeacher(BB3TeacherMixin, wow_srm):
    pass


class WoiSearchDialogueTeacher(BB3TeacherMixin, woi_srm):
    pass


class MsMarcoSearchDialogueTeacher(BB3TeacherMixin, msmarco_srm):
    pass


class FitsSearchDialogueTeacher(BB3TeacherMixin, FitsDialogueTeacher):
    def __init__(self, opt, shared=None):
        mutators = '+'.join(
            [
                'flatten_gold_human_mutator',
                'knowledge_appended_mutator',
                'skip_retrieval_mutator',
                'no_woi_gold_docs_mutator',
                'fits_pop_keys_mutator',
            ]
        )
        if opt.get('mutators'):
            mutators = '+'.join([mutators, opt['mutators']])
        logging.warning(f'overriding mutators to {mutators}')
        opt['mutators'] = mutators
        super().__init__(opt, shared)


class FunpediaSearchDialogueTeacher(BB3TeacherMixin, funpedia.DefaultTeacher):
    def __init__(self, opt, shared=None):
        mutators = '+'.join(['funpedia_to_bb3_mutator', 'skip_retrieval_mutator'])
        if opt.get('mutators'):
            mutators = '+'.join([mutators, opt['mutators']])
        logging.warning(f'overriding mutators to {mutators}')
        opt['mutators'] = mutators
        super().__init__(opt, shared)


class GoogleSgdSearchDialogueTeacher(BB3TeacherMixin, google_sgd.SystemTeacher):
    def __init__(self, opt, shared=None):
        mutators = '+'.join(['tod_to_srm_mutator', 'skip_retrieval_mutator'])
        if opt.get('mutators'):
            mutators = '+'.join([mutators, opt['mutators']])
        logging.warning(f'overriding mutators to {mutators}')
        opt['mutators'] = mutators
        super().__init__(opt, shared)


class TaskmasterSearchDialogueTeacher(BB3TeacherMixin, taskmaster.SystemTeacher):
    def __init__(self, opt, shared=None):
        mutators = '+'.join(['tod_to_srm_mutator', 'skip_retrieval_mutator'])
        if opt.get('mutators'):
            mutators = '+'.join([mutators, opt['mutators']])
        logging.warning(f'overriding mutators to {mutators}')
        opt['mutators'] = mutators
        super().__init__(opt, shared)


class Taskmaster2SearchDialogueTeacher(BB3TeacherMixin, taskmaster2.SystemTeacher):
    def __init__(self, opt, shared=None):
        mutators = '+'.join(['tod_to_srm_mutator', 'skip_retrieval_mutator'])
        if opt.get('mutators'):
            mutators = '+'.join([mutators, opt['mutators']])
        logging.warning(f'overriding mutators to {mutators}')
        opt['mutators'] = mutators
        super().__init__(opt, shared)


class Taskmaster3SearchDialogueTeacher(BB3TeacherMixin, taskmaster3.SystemTeacher):
    def __init__(self, opt, shared=None):
        mutators = '+'.join(['tod_to_srm_mutator', 'skip_retrieval_mutator'])
        if opt.get('mutators'):
            mutators = '+'.join([mutators, opt['mutators']])
        logging.warning(f'overriding mutators to {mutators}')
        opt['mutators'] = mutators
        super().__init__(opt, shared)


###################
# Entity Dialogue #
###################


def get_entity_dialogue_task_mutators(opt: Opt) -> str:
    """
    Set the mutators appropriately for the dialogue tasks.
    """
    mutators = '+'.join(
        ['flatten', 'extract_entity_for_response_model_bb3', 'skip_retrieval_mutator']
    )
    if opt.get('mutators'):
        mutators = '+'.join([mutators, opt['mutators']])
    logging.warning(f'overriding mutators to {mutators}')
    return mutators


class Convai2EntityDialogueTeacher(BB3TeacherMixin, convai2.NormalizedTeacher):
    def __init__(self, opt, shared=None):
        opt['mutators'] = get_entity_dialogue_task_mutators(opt)
        opt['task'] += ':no_cands'
        super().__init__(opt, shared)


class EDEntityDialogueTeacher(BB3TeacherMixin, ed.DefaultTeacher):
    def __init__(self, opt, shared=None):
        opt['mutators'] = get_entity_dialogue_task_mutators(opt)
        super().__init__(opt, shared)


class BSTEntityDialogueTeacher(BB3TeacherMixin, bst.DefaultTeacher):
    def __init__(self, opt, shared=None):
        opt['mutators'] = get_entity_dialogue_task_mutators(opt)
        super().__init__(opt, shared)


class MSCEntityDialogueTeacher(BB3TeacherMixin, msc.DefaultTeacher):
    def __init__(self, opt, shared=None):
        opt['mutators'] = get_entity_dialogue_task_mutators(opt)
        opt['include_session1'] = False
        super().__init__(opt, shared)


###################
# Memory Dialogue #
###################


def get_memory_dialogue_task_mutator(opt: Opt) -> str:
    """
    Set the mutators appropriately for the dialogue tasks.
    """
    mutators = '+'.join(
        [
            'convert_mkm_to_mrm_mutator',
            'skip_retrieval_mutator',
        ]
    )
    assert opt.get('mutators')
    mutators = '+'.join([opt['mutators'], mutators])
    logging.warning(f'overriding mutators to {mutators}')
    return mutators


class Convai2MemoryDialogueFromPersOverlapTeacher(
    BB3TeacherMixin, convai2.NormalizedTeacher
):
    def __init__(self, opt, shared=None):
        opt['mutators'] = get_memory_knowledge_pers_overlap_task_mutator(opt)
        opt['mutators'] = get_memory_dialogue_task_mutator(opt)
        opt['task'] += ':no_cands'
        super().__init__(opt, shared)


class EDMemoryDialogueFromPersOverlapTeacher(BB3TeacherMixin, ed.DefaultTeacher):
    def __init__(self, opt, shared=None):
        opt['mutators'] = get_memory_knowledge_pers_overlap_task_mutator(opt)
        mutators = get_memory_dialogue_task_mutator(opt)
        opt['mutators'] = mutators
        super().__init__(opt, shared)


class BSTMemoryDialogueFromPersOverlapTeacher(BB3TeacherMixin, bst.DefaultTeacher):
    def __init__(self, opt, shared=None):
        opt['mutators'] = get_memory_knowledge_pers_overlap_task_mutator(opt)
        opt['mutators'] = get_memory_dialogue_task_mutator(opt)
        super().__init__(opt, shared)


class MSCMemoryDialogueFromPersOverlapTeacher(BB3TeacherMixin, msc.DefaultTeacher):
    def __init__(self, opt, shared=None):
        opt['mutators'] = get_memory_knowledge_pers_overlap_task_mutator(opt)
        opt['mutators'] = get_memory_dialogue_task_mutator(opt)
        opt['include_session1'] = False
        super().__init__(opt, shared)


class Convai2MemoryDialogueFromUttOverlapTeacher(
    BB3TeacherMixin, convai2.NormalizedTeacher
):
    def __init__(self, opt, shared=None):
        opt['mutators'] = get_memory_knowledge_utt_overlap_task_mutator(opt)
        opt['mutators'] = get_memory_dialogue_task_mutator(opt)
        opt['task'] += ':no_cands'
        super().__init__(opt, shared)


class EDMemoryDialogueFromUttOverlapTeacher(BB3TeacherMixin, ed.DefaultTeacher):
    def __init__(self, opt, shared=None):
        opt['mutators'] = get_memory_knowledge_utt_overlap_task_mutator(opt)
        opt['mutators'] = get_memory_dialogue_task_mutator(opt)
        super().__init__(opt, shared)


class BSTMemoryDialogueFromUttOverlapTeacher(BB3TeacherMixin, bst.DefaultTeacher):
    def __init__(self, opt, shared=None):
        opt['mutators'] = get_memory_knowledge_utt_overlap_task_mutator(opt)
        opt['mutators'] = get_memory_dialogue_task_mutator(opt)
        super().__init__(opt, shared)


class MSCMemoryDialogueFromUttOverlapTeacher(BB3TeacherMixin, msc.DefaultTeacher):
    def __init__(self, opt, shared=None):
        opt['mutators'] = get_memory_knowledge_utt_overlap_task_mutator(opt)
        opt['mutators'] = get_memory_dialogue_task_mutator(opt)
        opt['include_session1'] = False
        super().__init__(opt, shared)


####################
# Vanilla Dialogue #
####################


class SaferdialoguesVanillaDialogueTeacher(
    BB3TeacherMixin, Saferdialogues.DefaultTeacher
):
    def __init__(self, opt, shared=None):
        mutators = '+'.join(['flatten', 'skip_retrieval_mutator'])
        if opt.get('mutators'):
            mutators = '+'.join([mutators, opt['mutators']])
        logging.warning(f'overriding mutators to {mutators}')
        opt['mutators'] = mutators
        super().__init__(opt, shared)


class WowVanillaDialogueTeacher(BB3TeacherMixin, wow.DefaultTeacher):
    def __init__(self, opt, shared=None):
        mutators = '+'.join(
            [
                'flatten',
                'wow_only_no_passage_used',
                'wow_to_woi',
                'woi_pop_documents_mutator',
                'skip_retrieval_mutator',
            ]
        )
        if opt.get('mutators'):
            mutators = '+'.join([mutators, opt['mutators']])
        logging.warning(f'overriding mutators to {mutators}')
        opt['mutators'] = mutators
        opt['add_missing_turns'] = 'all'
        super().__init__(opt, shared)


class WoiVanillaDialogueTeacher(BB3TeacherMixin, woi.DefaultTeacher):
    def __init__(self, opt, shared=None):
        mutators = '+'.join(
            [
                'flatten',
                'woi_keep_only_no_passage_used',
                'skip_retrieval_mutator',
                'woi_pop_documents_mutator',
            ]
        )
        if opt.get('mutators'):
            mutators = '+'.join([mutators, opt['mutators']])
        logging.warning(f'overriding mutators to {mutators}')
        opt['mutators'] = mutators
        super().__init__(opt, shared)


class LightVanillaDialogueTeacher(BB3TeacherMixin, light.SimpleMultiTeacher):
    def __init__(self, opt, shared=None):
        opt['light_label'] = 'speech'
        mutators = '+'.join(['flatten', 'skip_retrieval_mutator'])
        if opt.get('mutators'):
            mutators = '+'.join([mutators, opt['mutators']])
        logging.warning(f'overriding mutators to {mutators}')
        opt['mutators'] = mutators
        super().__init__(opt, shared)


class LightWildVanillaDialogueTeacher(BB3TeacherMixin, light_wild.SimpleMultiTeacher):
    def __init__(self, opt, shared=None):
        opt['light_label'] = 'speech'
        mutators = '+'.join(['flatten', 'skip_retrieval_mutator'])
        if opt.get('mutators'):
            mutators = '+'.join([mutators, opt['mutators']])
        logging.warning(f'overriding mutators to {mutators}')
        opt['mutators'] = mutators
        super().__init__(opt, shared)


class EDVanillaDialogueTeacher(BB3TeacherMixin, ed.DefaultTeacher):
    def __init__(self, opt, shared=None):
        mutators = '+'.join(['flatten', 'skip_retrieval_mutator'])
        if opt.get('mutators'):
            mutators = '+'.join([mutators, opt['mutators']])
        logging.warning(f'overriding mutators to {mutators}')
        opt['mutators'] = mutators
        super().__init__(opt, shared)


class Convai2VanillaDialogueTeacher(BB3TeacherMixin, convai2.NormalizedTeacher):
    def __init__(self, opt, shared=None):
        mutators = '+'.join(['flatten', 'skip_retrieval_mutator'])
        if opt.get('mutators'):
            mutators = '+'.join([mutators, opt['mutators']])
        logging.warning(f'overriding mutators to {mutators}')
        opt['mutators'] = mutators
        opt['task'] += ':no_cands'
        super().__init__(opt, shared)


class MSCVanillaDialogueTeacher(BB3TeacherMixin, msc.DefaultTeacher):
    def __init__(self, opt, shared=None):
        mutators = '+'.join(['flatten', 'skip_retrieval_mutator'])
        if opt.get('mutators'):
            mutators = '+'.join([mutators, opt['mutators']])
        logging.warning(f'overriding mutators to {mutators}')
        opt['mutators'] = mutators
        opt['include_session1'] = False
        super().__init__(opt, shared)


#####################
# Grounded Dialogue #
#####################


class Convai2StyleGroundingDialogueTeacher(
    BB3TeacherMixin, style_gen.LabeledConvAI2PersonaTopicifierTeacher
):
    def __init__(self, opt, shared=None):
        mutators = '+'.join(['flatten', 'style_gen_to_grm', 'skip_retrieval_mutator'])
        if opt.get('mutators'):
            mutators = '+'.join([mutators, opt['mutators']])
        logging.warning(f'overriding mutators to {mutators}')
        opt['mutators'] = mutators
        opt['include_session1'] = False
        super().__init__(opt, shared)


class BSTStyleGroundingDialogueTeacher(
    BB3TeacherMixin, style_gen.LabeledBlendedSkillTalkTeacher
):
    def __init__(self, opt, shared=None):
        mutators = '+'.join(['flatten', 'style_gen_to_grm', 'skip_retrieval_mutator'])
        if opt.get('mutators'):
            mutators = '+'.join([mutators, opt['mutators']])
        logging.warning(f'overriding mutators to {mutators}')
        opt['mutators'] = mutators
        opt['include_session1'] = False
        super().__init__(opt, shared)



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
"""
Distributed script for running TOD model-model chats.

Not to be called directly; should be called from SLURM
"""

from parlai.scripts.tod_world_script import TodWorldScript
from parlai.core.script import ParlaiScript
import parlai.utils.distributed as distributed_utils


class DistributedTodWorldScript(ParlaiScript):
    @classmethod
    def setup_args(cls):
        parser = TodWorldScript.setup_args()
        parser.add_distributed_training_args()
        parser.add_argument("--port", type=int, default=61337, help="TCP port number")
        return parser

    def run(self):
        with distributed_utils.slurm_distributed_context(self.opt) as opt:
            return TodWorldScript(opt).run()


if __name__ == "__main__":
    DistributedTodWorldScript.main()



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
"""
Metrics handlers - ie, classes that handle generations from Tod World and calculates metrics from them.

Note that only metrics handler classes in `WORLD_METRIC_HANDLERS` of ` parlai/core/tod/world_metrics_handlers.py` are actively being recorded as metrics. These "extended" metrics were ones we experimented with at one point in the past, found to be inconclusive, and are including primarily to not delete already done work. (Or for testing purposes.)
"""

from parlai.core.message import Message
from parlai.core.metrics import (
    Metric,
    AverageMetric,
    normalize_answer,
    BleuMetric,
    SumMetric,
)
from parlai.core.tod.tod_core import (
    STANDARD_API_NAME_SLOT,
    STANDARD_REQUIRED_KEY,
    STANDARD_OPTIONAL_KEY,
    STANDARD_API_SCHEMAS,
)
from typing import Dict, List, Optional, Tuple
from parlai.core.tod.world_metrics_handlers import (
    TodMetricsHandler,
    register_metrics_handler,
    _ApiCallGoalInteractionHelper,
    goals_hit_helper,
)

try:
    from nltk.translate import bleu_score as nltkbleu
except ImportError:
    # User doesn't have nltk installed, so we can't use it for bleu
    # We'll just turn off things, but we might want to warn the user
    nltkbleu = None

################################
# Functions and classes associated with calculating statistics between API Calls and Goals.


def get_req_only_goals(goals_list: List[Dict], api_schemas: List[Dict]) -> List[Dict]:
    """
    Given a list of goals and a list of api schemas that say if slots are required or
    optional, this function filters for the goals to be only the required ones.

    If we have no api schemas or a goal is malformed, we return the empty list. If a
    goal is malformed, we print a warning, since this whole req-only goals thing is
    experimental at best anyhow.
    """
    if len(api_schemas) == 0:
        return []
    result = []
    for goal in goals_list:
        req_goals = {}
        method = goal.get(STANDARD_API_NAME_SLOT, None)
        if method is None:
            return []
        required = []
        for schema in api_schemas:
            if schema.get(STANDARD_API_NAME_SLOT, "") == method:
                required = schema.get(STANDARD_REQUIRED_KEY, {})
        print("-".join(required))
        for key in required:
            if key not in goal:
                print(f"No required key `{key}` in goal `{goal}`")
                return []
            req_goals[key] = goal[key]
        if len(req_goals) > 0:
            req_goals[STANDARD_API_NAME_SLOT] = method  # for consistency with all.
            result.append(req_goals)
    return result


def goals_slots_helper(
    goals: List[Dict], turnDict: List[Dict]
) -> Tuple[Tuple[int, int], Tuple[int, int]]:
    """
    Helper function to see how well the slot keys + slot values match between attempted
    API calls and goals.

    Output is precision, recall.
    """
    all_call_slots = {k: v for call in turnDict for k, v in call.items()}
    all_goal_slots = {k: v for goal in goals for k, v in goal.items()}
    goal_in_call = {
        k: v
        for k, v in all_call_slots.items()
        if all_goal_slots.get(k, "definitelyNotInValuexyz") == v
    }
    call_in_goal = {
        k: v
        for k, v in all_goal_slots.items()
        if all_call_slots.get(k, "definitelyNotInValuexyz") == v
    }

    print(goal_in_call, all_call_slots)

    return (
        AverageMetric(len(goal_in_call), len(all_call_slots)),
        AverageMetric(len(call_in_goal), len(all_goal_slots)),
    )


@register_metrics_handler
class LegacyGoalApiCallInteractionsMetricsHandler(_ApiCallGoalInteractionHelper):
    """
    This class was reporting a few too many metrics, but is useful for test purposes, so
    we're keeping it around.

    `AllGoalApiCallSuccessMetricsHandler` is the streamlined, less spammy version of
    this class.
    """

    def handle_goals(
        self, message: Message, goals: List[Dict]
    ) -> Optional[Dict[str, Metric]]:
        self.goals = goals
        self.required_goals = get_req_only_goals(goals, self.api_schemas)

    def get_episode_metrics(self) -> Optional[Dict[str, Metric]]:
        all_goals_hit, all_goals_hit_turn_count, all_part_hit = goals_hit_helper(
            self.goals, self.api_turns
        )
        all_precision, all_recall = goals_slots_helper(self.goals, self.api_turns)
        req_goals_hit, req_goals_hit_turn_count, req_part_hit = goals_hit_helper(
            self.required_goals, self.api_turns, permissive=True
        )
        req_precision, req_recall = goals_slots_helper(
            self.required_goals, self.api_turns
        )
        call_attempts = len(self.api_turns)
        return {
            "all_goals_hit": all_goals_hit,
            "all_goals_hit_turn_count": all_goals_hit_turn_count,
            "all_goals_fractional_hit": all_part_hit,
            "all_goals_slot_precision": all_precision,
            "all_goals_slot_recall": all_recall,
            "req_goals_hit": req_goals_hit,
            "req_goals_hit_turn_count": req_goals_hit_turn_count,
            "req_goals_fractional_hit": req_part_hit,
            "req_goals_slot_precision": req_precision,
            "req_goals_slot_recall": req_recall,
            "call_attempts": AverageMetric(call_attempts),
        }


@register_metrics_handler
class UserGoalSlotCoverageMetricHandler(TodMetricsHandler):
    """
    How well does our user simulator do at outputting utterances that goes closer to
    satisfying relevant groundinged goals? Does it dump out all of the slots at once or
    is it more intelligent than that?

    Since this is the user and we don't know the identity of potential slots, we ignore
    the short (< 4 chars) goal slots since this tends to be things that are substrings
    of other things. (Ex. "2" showing up as # of people in a reservation, but also
    showing up as a phone number.)
    """

    def episode_reset(self):
        self.mentioned_all_slot_values = set()
        self.mentioned_req_slot_values = set()
        self.all_goal_slot_values = set()
        self.all_req_goal_slot_values = set()

    def handle_goals(
        self, message: Message, goals: List[Dict]
    ) -> Optional[Dict[str, Metric]]:
        """
        Parse out all the slots as a blob, filtering out for short things.
        """
        required_goals = get_req_only_goals(goals, self.api_schemas)

        def get_slot_values(goal_list):
            result = set()
            for goal in goal_list:
                for key, value in goal.items():
                    if key is not STANDARD_API_NAME_SLOT and len(value) > 3:
                        result.add(value)
            return result

        self.all_goal_slot_values = get_slot_values(goals)
        self.all_req_goal_slot_values = get_slot_values(required_goals)

    def handle_user_utt(
        self, message: Message, prefix_stripped_text: str
    ) -> Optional[Dict[str, Metric]]:
        """
        Grab slots out of the user utterance based on an exact match.
        """
        utterance = prefix_stripped_text

        def get_slots(utt, options):
            results = set()
            for option in options:
                if option in utt:
                    results.add(option)
            return results

        all_slot_values_here = get_slots(utterance, self.all_goal_slot_values)
        req_slot_values_here = get_slots(utterance, self.all_req_goal_slot_values)

        self.mentioned_all_slot_values |= all_slot_values_here
        self.mentioned_req_slot_values |= req_slot_values_here

        metrics = {}
        metrics["user_utt_avg_any_slot"] = AverageMetric(len(all_slot_values_here))
        metrics["user_utt_avg_req_slot"] = AverageMetric(len(req_slot_values_here))
        return metrics

    def get_episode_metrics(self) -> Optional[Dict[str, Metric]]:
        result = {
            "user_any_goal_slots_recall": AverageMetric(
                len(self.mentioned_all_slot_values), len(self.all_goal_slot_values)
            ),
            "user_req_goal_slots_recall": AverageMetric(
                len(self.mentioned_req_slot_values), len(self.all_req_goal_slot_values)
            ),
        }

        self.mentioned_all_slot_values = set()
        self.mentioned_req_slot_values = set()
        return result


class _ExactRepeatMetricsHandler(TodMetricsHandler):
    """
    Helper class for defining % of episodes where a given agent type has exactly
    repeated the same utterance.
    """

    def episode_reset(self):
        self.turns = []
        self.repeated = False

    def metric_key(self):
        raise NotImplementedError("must implement")

    def handle_message_helper(
        self, prefix_stripped_text: str
    ) -> Optional[Dict[str, Metric]]:
        normalized = normalize_answer(prefix_stripped_text)
        if normalized in self.turns:
            self.repeated = True
        self.turns.append(normalized)

    def get_episode_metrics(self) -> Optional[Dict[str, Metric]]:
        repeat = int(self.repeated)
        self.repeated = False
        self.turns = []
        return {self.metric_key(): AverageMetric(repeat)}


@register_metrics_handler
class UserUttRepeatMetricHandler(_ExactRepeatMetricsHandler):
    def metric_key(self):
        return "user_utt_repeat"

    def handle_user_utt(
        self, message: Message, prefix_stripped_text: str
    ) -> Optional[Dict[str, Metric]]:
        return self.handle_message_helper(prefix_stripped_text)


@register_metrics_handler
class SystemUttRepeatMetricHandler(_ExactRepeatMetricsHandler):
    def metric_key(self):
        return "sys_utt_repeat"

    def handle_sys_utt(
        self, message: Message, prefix_stripped_text: str
    ) -> Optional[Dict[str, Metric]]:
        return self.handle_message_helper(prefix_stripped_text)


@register_metrics_handler
class _Bleu3MetricsHandler(TodMetricsHandler):
    """
    For a given agent, this calculates the Bleu-3 of a new turn against prior turns.

    This is an alternate metric for repetativeness
    """

    def episode_reset(self):
        self.turns = []

    def metric_key(self):
        raise NotImplementedError("must implement")

    def handle_message_helper(
        self, prefix_stripped_text: str
    ) -> Optional[Dict[str, Metric]]:
        here = [normalize_answer(x) for x in prefix_stripped_text.split(" ")]
        score = 1
        if len(self.turns) > 0:
            score = nltkbleu.corpus_bleu(
                [self.turns],
                [here],
                smoothing_function=nltkbleu.SmoothingFunction(epsilon=1e-12).method1,
                weights=[1.0 / 3.0] * 3,
            )
        self.turns.append(here)
        return {self.metric_key(): BleuMetric(score)}


@register_metrics_handler
class UserUttSelfBleu3MetricHandler(_Bleu3MetricsHandler):
    def metric_key(self):
        return "user_utt_self_bleu3"

    def handle_user_utt(
        self, message: Message, prefix_stripped_text: str
    ) -> Optional[Dict[str, Metric]]:
        return self.handle_message_helper(prefix_stripped_text)


@register_metrics_handler
class SystemUttSelfBleu3MetricHandler(_Bleu3MetricsHandler):
    def metric_key(self):
        return "sys_utt_self_bleu3"

    def handle_sys_utt(
        self, message: Message, prefix_stripped_text: str
    ) -> Optional[Dict[str, Metric]]:
        return self.handle_message_helper(prefix_stripped_text)


@register_metrics_handler
class ApiCallMalformedMetricHandler(TodMetricsHandler):
    def episode_reset(self):
        self.api_schemas = []

    def handle_api_call(
        self, message: Message, api_call: Dict
    ) -> Optional[Dict[str, Metric]]:
        if STANDARD_API_SCHEMAS in message["text"]:
            return  # Happens for API call groundingion, so it's fine
        if len(api_call) == 0:
            return
        if STANDARD_API_NAME_SLOT not in api_call:
            return {
                "apiCall_wellFormed": AverageMetric(0),
                "apiCall_hasSlotsButNoApiNameSlot_count": SumMetric(1),
            }
        method = api_call[STANDARD_API_NAME_SLOT]

        method_found = False
        if len(self.api_schemas) > 0:
            for schema in self.api_schemas:
                if method == schema.get(STANDARD_API_NAME_SLOT, ""):
                    method_found = True
                    check = api_call.keys()
                    required = set(schema.get(STANDARD_REQUIRED_KEY, []))
                    required.add(STANDARD_API_NAME_SLOT)
                    for req in required:
                        if req not in check:  # miissing required
                            return {
                                "apiCall_wellFormed": AverageMetric(0),
                                "apiCall_missingRequiredSlot_count": SumMetric(1),
                            }
                    opt_count = 0
                    for opt in schema.get(STANDARD_OPTIONAL_KEY, []):
                        if opt in check:
                            opt_count += 1
                    if opt_count + len(required) != len(check):
                        # have extra APIs that are not
                        return {
                            "apiCall_wellFormed": AverageMetric(0),
                            "apiCall_hasExtraParams_count": SumMetric(1),
                        }
                    break
        if method_found:
            return {
                "apiCall_wellFormed": AverageMetric(1),
                "apiCall_wellFormed_count": SumMetric(1),
            }
        return {
            "apiCall_wellFormed": AverageMetric(0),
            "apiCall_methodDNE_count": SumMetric(1),
        }


@register_metrics_handler
class PseudoInformMetricsHandler(TodMetricsHandler):
    """
    Pseudo-inform rate.
    """

    def episode_reset(self):
        self.api_resp_slots = {}

    def handle_api_resp(
        self, message: Message, api_resp: Dict
    ) -> Optional[Dict[str, Metric]]:
        self.api_resp_slots.update(api_resp)

    def handle_sys_utt(
        self, message: Message, prefix_stripped_text: str
    ) -> Optional[Dict[str, Metric]]:
        count = 0
        for val in self.api_resp_slots.values():
            if val in prefix_stripped_text:
                count += 1
        result = {"pseudo_inform_allSysTurns": AverageMetric(count)}
        if len(self.api_resp_slots) > 0:
            result["pseudo_inform_postApiRespSysTurns"] = AverageMetric(count)
        return result



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
"""
SeeKeR Dialogue Tasks.
"""
from typing import Optional

from parlai.core.opt import Opt
from parlai.core.params import ParlaiParser
from parlai.core.teachers import MultiTaskTeacher
import parlai.tasks.convai2.agents as convai2
import parlai.tasks.blended_skill_talk.agents as bst
import parlai.tasks.empathetic_dialogues.agents as ed
import parlai.tasks.wizard_of_internet.agents as woi
import parlai.tasks.wizard_of_wikipedia.agents as wow
import parlai.tasks.msc.agents as msc
import parlai.tasks.ms_marco.agents as ms_marco

import parlai.utils.logging as logging

import projects.seeker.tasks.mutators  # type: ignore   # noqa: F401


class WoiDialogueTeacher(woi.DefaultTeacher):
    def __init__(self, opt, shared=None):
        mutators = '+'.join(
            [
                'flatten',
                'woi_pop_documents_mutator',
                'woi_filter_no_passage_used',
                'woi_add_checked_sentence_to_input',
                'skip_retrieval_mutator',
            ]
        )
        if opt.get('mutators'):
            mutators = '+'.join([mutators, opt['mutators']])
        logging.warning(f'overriding mutators to {mutators}')
        opt['mutators'] = mutators
        super().__init__(opt, shared)
        self.id = "WoiDialogueTeacher"


class WowDialogueTeacher(wow.DefaultTeacher):
    def __init__(self, opt, shared=None):
        opt['add_missing_turns'] = 'all'
        mutators = '+'.join(
            [
                'flatten',
                'wow_filter_no_passage_used',
                'wow_add_checked_sentence_to_input',
                'skip_retrieval_mutator',
                'wow_to_woi',
                'woi_pop_documents_mutator',
            ]
        )
        if opt.get('mutators'):
            mutators = '+'.join([mutators, opt['mutators']])
        logging.warning(f'overriding mutators to {mutators}')
        opt['mutators'] = mutators
        super().__init__(opt, shared)
        self.id = "WowDialogueTeacher"


class MsMarcoDialogueTeacher(ms_marco.DefaultTeacher):
    def __init__(self, opt, shared=None):
        mutators = '+'.join(
            [
                'ms_marco_filter_has_answer',
                'ms_marco_create_fid_docs',
                'ms_marco_find_selected_sentence_for_response',
                'woi_pop_documents_mutator',
                'skip_retrieval_mutator',
            ]
        )
        if opt.get('mutators'):
            mutators = '+'.join([mutators, opt['mutators']])
        logging.warning(f'overriding mutators to {mutators}')
        opt['mutators'] = mutators
        super().__init__(opt, shared)
        self.id = "MsMarcoDialogueTeacher"


def get_dialogue_task_mutators(opt: Opt) -> str:
    """
    Set the mutators appropriately for the dialogue tasks.
    """
    mutators = '+'.join(
        ['flatten', 'extract_entity_for_response_model', 'skip_retrieval_mutator']
    )
    if opt.get('mutators'):
        mutators = '+'.join([mutators, opt['mutators']])
    logging.warning(f'overriding mutators to {mutators}')
    return mutators


class Convai2DialogueTeacher(convai2.NormalizedTeacher):
    def __init__(self, opt, shared=None):
        opt['mutators'] = get_dialogue_task_mutators(opt)
        opt['task'] += ':no_cands'
        super().__init__(opt, shared)
        self.id = 'Convai2DialogueTeacher'


class EDDialogueTeacher(ed.DefaultTeacher):
    def __init__(self, opt, shared=None):
        opt['mutators'] = get_dialogue_task_mutators(opt)
        super().__init__(opt, shared)
        self.id = 'EDDialogueTeacher'


class BSTDialogueTeacher(bst.DefaultTeacher):
    def __init__(self, opt, shared=None):
        opt['mutators'] = get_dialogue_task_mutators(opt)
        super().__init__(opt, shared)
        self.id = 'BSTDialogueTeacher'


class MSCDialogueTeacher(msc.DefaultTeacher):
    def __init__(self, opt, shared=None):
        opt['mutators'] = get_dialogue_task_mutators(opt)
        opt['include_session1'] = False
        super().__init__(opt, shared)
        self.id = 'MSCDialogueTeacher'


class MSCDialogueOverlapTeacher(msc.DefaultTeacher):
    def __init__(self, opt, shared=None):
        opt['mutators'] = '+'.join(
            ['flatten', 'msc_find_selected_sentence_response', 'skip_retrieval_mutator']
        )
        opt['include_session1'] = False
        super().__init__(opt, shared)
        self.id = 'MSCDialogueOverlapTeacher'


class DialogueTeacher(MultiTaskTeacher):
    @classmethod
    def add_cmdline_args(
        cls, parser: ParlaiParser, partial_opt: Optional[Opt] = None
    ) -> ParlaiParser:
        WoiDialogueTeacher.add_cmdline_args(parser, partial_opt)
        WowDialogueTeacher.add_cmdline_args(parser, partial_opt)
        MsMarcoDialogueTeacher.add_cmdline_args(parser, partial_opt)
        Convai2DialogueTeacher.add_cmdline_args(parser, partial_opt)
        EDDialogueTeacher.add_cmdline_args(parser, partial_opt)
        BSTDialogueTeacher.add_cmdline_args(parser, partial_opt)
        MSCDialogueTeacher.add_cmdline_args(parser, partial_opt)
        MSCDialogueOverlapTeacher.add_cmdline_args(parser, partial_opt)
        return parser

    def __init__(self, opt, shared=None):
        tasks = [
            f"projects.seeker.tasks.dialogue:{teacher}"
            for teacher in [
                'WoiDialogueTeacher',
                'WowDialogueTeacher',
                'MsMarcoDialogueTeacher',
                'Convai2DialogueTeacher',
                'EDDialogueTeacher',
                'BSTDialogueTeacher',
                'MSCDialogueTeacher',
                'MSCDialogueOverlapTeacher',
            ]
        ]
        opt['task'] = ','.join(tasks)
        super().__init__(opt, shared)


class DefaultTeacher(DialogueTeacher):
    pass



#!/usr/bin/env python3
# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from enum import Enum


class FeedbackType(Enum):
    PERFECT = "perfect"
    CORRECT_SEARCH_QUERY = "better_search_query"
    CORRECT_DOC = "better_search_doc"
    CORRECT_HISTORY = "better_history"
    CORRECT_RESPONSE = "better_others"
    NONE = 'none'


NO_SEARCH = '__NO__SEARCH__'
TITLE_PASSAGE_DELIMITER = ' | '

OK_LABEL = '__ok__'
NOT_OK_LABEL = '__notok__'

UNWANTED_TOKENS = ['_potentially_unsafe__', '_pots and pots_']



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
"""
SeeKeR Agent.

This agent enables a four stage process:

1) determine if search
2) generate search query
3) generate knowledge response
4) generate dialogue response
"""
from collections import defaultdict
import copy
import torch
from types import MethodType
from typing import List, Tuple, Optional, Dict, Any

from parlai.agents.bart.bart import BartAgent
from parlai.agents.fid.fid import (
    FidAgent,
    GoldDocRetrieverFiDAgent,
    SearchQuerySearchEngineFiDAgent,
    WizIntGoldDocRetrieverFiDAgent,
)
from parlai.agents.rag.args import setup_rag_args
from parlai.agents.rag.retrievers import Document
from parlai.core.agents import Agent, create_agent_from_shared, create_agent
from parlai.core.message import Message
from parlai.core.mutators import MessageMutator, Mutator
from parlai.core.params import ParlaiParser
from parlai.core.opt import Opt
from parlai.core.torch_agent import Batch, Output, History
from parlai.core.torch_generator_agent import TorchGeneratorAgent
from parlai.tasks.wizard_of_wikipedia.agents import TOKEN_KNOWLEDGE, TOKEN_END_KNOWLEDGE
import parlai.utils.logging as logging

from projects.seeker.agents.seeker_modules import ComboFidModel
from projects.seeker.utils import (
    SearchDecision,
    drm_get_batch_context,
    krm_get_batch_context,
    krm_get_batch_context_only_knowledge,
    GENERATE_QUERY,
    IS_SEARCH_REQUIRED,
    DO_SEARCH,
    DO_NOT_SEARCH,
)
from projects.seeker.agents.modular_agent import ModularAgentMixin


class ComboFidAgent(FidAgent):
    """
    The ComboFidAgent is used to *train* a SeeKeR model.

    This agent operates as a standard FiD model when there are retrieved documents.

    Otherwise, it operates as a standard transformer/generator.
    """

    @classmethod
    def add_cmdline_args(
        cls, parser: ParlaiParser, partial_opt: Optional[Opt] = None
    ) -> ParlaiParser:
        super().add_cmdline_args(parser, partial_opt)
        combo_fid = parser.add_argument_group('Combo Fid Group')
        combo_fid.add_argument(
            '--skip-retrieval-key',
            type=str,
            default='skip_retrieval',
            help='key in observation determining whether to skip retrieval.',
        )
        combo_fid.add_argument(
            '--serializable',
            type='bool',
            default=False,
            help='Whether to make model act output fully serializable.',
        )
        combo_fid.add_argument(
            '--force-skip-retrieval',
            type='bool',
            default=False,
            help='If True, we force skip retrieval on any/all incoming examples',
        )

    def build_model(self) -> ComboFidModel:
        """
        Build and return ComboFidModel.
        """
        if self.generation_model == 't5':
            raise RuntimeError('T5 currently not supported')
        else:
            model = ComboFidModel(self.opt, self.dict)
        if self.opt['embedding_type'] != 'random':
            self._copy_embeddings(
                model.encoder.embeddings.weight, self.opt['embedding_type']
            )
        return model

    def batchify(self, obs_batch: List[Message], sort: bool = False) -> Batch:
        """
        Overrides FidAgent.batchify to add skip retrieval input vec.

        Additionally adds the prior knowledge responses to the batch. This allows
        context blocking in a KRM setup that includes the prior generations from the
        knowledge component.
        """
        batch = super().batchify(obs_batch, sort)
        valid_exs = [ex for ex in obs_batch if self.is_valid(ex)]
        if valid_exs:
            if self.opt.get('force_skip_retrieval', False):
                skip_retrieval = [True] * len(valid_exs)
            else:
                skip_retrieval = [
                    ex.get(self.opt['skip_retrieval_key'], False) for ex in valid_exs
                ]
            batch.skip_retrieval_vec = torch.BoolTensor(skip_retrieval)
            if any(ex.get('prior_knowledge_responses') for ex in valid_exs):
                vecs, _lens = self._pad_tensor(
                    [
                        self.dict.txt2vec(ex.get('prior_knowledge_responses'))
                        for ex in valid_exs
                    ]
                )
                batch.prior_knowledge_responses_vec = vecs
        return batch

    def _model_input(
        self, batch: Batch
    ) -> Tuple[
        torch.LongTensor,
        torch.LongTensor,
        torch.LongTensor,
        torch.LongTensor,
        torch.BoolTensor,
    ]:
        """
        Override FidModel._model_input to add skip_retrieval_vec.
        """
        return (
            batch.text_vec,
            batch.text_vec.ne(self.NULL_IDX).sum(1),
            batch.query_vec,
            batch.input_turn_cnt_vec,
            batch.skip_retrieval_vec,
        )

    def get_retrieved_knowledge(self, message: Message) -> List[Document]:
        if message.get('skip_retrieval'):
            return []
        return super().get_retrieved_knowledge(message)

    def eval_step(self, batch: Batch) -> Optional[Output]:
        """
        Add top documents to the output.
        """
        output = TorchGeneratorAgent.eval_step(self, batch)
        if output is not None and not self.opt.get('serializable', False):
            output.top_docs = self.model_api.get_top_docs()
        return output


class ComboFidSearchQueryAgent(ComboFidAgent, SearchQuerySearchEngineFiDAgent):
    pass


class ComboFidGoldDocumentAgent(ComboFidAgent, WizIntGoldDocRetrieverFiDAgent):
    pass


class SeekerAgent(ModularAgentMixin):
    """
    SeeKeR Agent.

    One module performs search query generation, knowledge response generation,
    and dialogue response generation.

    Additionally, may also perform search decision generation.

    To provide arguments to a sub-agent, simply prepend the argument with one of
    `sdm/sqm/krm/drm`.
    """

    knowledge_agent: Agent
    dialogue_agent: Agent

    @classmethod
    def get_additional_agent_args(cls) -> ParlaiParser:
        """
        Return a parser with arguments sourced from several sub models.
        """
        additional_agent_parser = ParlaiParser(add_parlai_args=False)
        BartAgent.add_cmdline_args(additional_agent_parser)
        setup_rag_args(additional_agent_parser)
        GoldDocRetrieverFiDAgent.add_cmdline_args(additional_agent_parser)
        SearchQuerySearchEngineFiDAgent.add_cmdline_args(additional_agent_parser)
        WizIntGoldDocRetrieverFiDAgent.add_cmdline_args(additional_agent_parser)
        ComboFidAgent.add_cmdline_args(additional_agent_parser)
        return additional_agent_parser

    @classmethod
    def add_additional_subagent_args(cls, parser: ParlaiParser) -> ParlaiParser:
        """
        Add additional args for the sub agents.
        """
        additional_agent_parser = cls.get_additional_agent_args()
        for action in additional_agent_parser._actions:
            key = max(action.option_strings, key=lambda x: len(x))
            type = action.type

            for prefix in ['krm', 'drm', 'sqm', 'sdm']:
                parser.add_argument(
                    f'--{prefix}-{key.strip("-")}', type=type, required=False
                )
        return parser

    @classmethod
    def add_cmdline_args(
        cls, parser: ParlaiParser, partial_opt: Optional[Opt] = None
    ) -> ParlaiParser:
        """
        Command line args for the Seeker Agent.
        """
        cls.add_additional_subagent_args(parser)
        group = parser.add_argument_group('SeeKeR Agent Args')
        group.add_argument(
            '--krm-model', type=str, help='agent (not model file) to load for krm'
        )
        group.add_argument(
            '--drm-model', type=str, help='agent (not model file) to load for krm'
        )
        group.add_argument(
            '--sqm-model', type=str, help='agent (not model file) to load for sqm'
        )
        group.add_argument(
            '--sdm-model', type=str, help='agent (not model file) to load for sdm'
        )
        group.add_argument(
            '--search-decision-control-token',
            type=str,
            default=IS_SEARCH_REQUIRED,
            help='control token for search decision model',
        )
        group.add_argument(
            '--search-decision-do-search-reply',
            type=str,
            default=DO_SEARCH,
            help='control token returned by SDM to indicate searching',
        )
        group.add_argument(
            '--search-decision-dont-search-reply',
            type=str,
            default=DO_NOT_SEARCH,
            help='control token returned by SDM to indicate not searching',
        )
        group.add_argument(
            '--search-query-control-token',
            type=str,
            default=GENERATE_QUERY,
            help='control token for search query model',
        )
        group.add_argument(
            '--knowledge-response-control-token',
            type=str,
            default=None,
            help='control token for knowledge response model',
        )
        group.add_argument(
            '--all-model-path',
            type=str,
            default=None,
            help='If specified, load all models with this path',
        )
        group.add_argument(
            '--search-decision',
            type=str,
            default=SearchDecision.ALWAYS.value,
            choices=[s.value for s in SearchDecision],
        )
        group.add_argument(
            '--min-knowledge-length-when-search',
            type=int,
            default=-1,
            help='Set the minimum generated knowledge length when searching.',
        )
        group.add_argument(
            '--beam-disregard-knowledge-for-context-blocking',
            type='bool',
            default=False,
            help='If True disregard the knowledge input for DRM context blocking.',
        )
        group.add_argument(
            '--include-knowledge-in-krm-context-blocking',
            type=bool,
            default=True,
            help='If True, put the KRM responses in the context for the KRM.',
        )
        group.add_argument(
            '--exclude-context-in-krm-context-blocking',
            type=bool,
            default=False,
            help='Used in conjunction with --include-knowledge-in-krm-context-blocking. '
            'If specified, only block on the knowledge, and not the concatenation.',
        )
        group.add_argument(
            '--inject-query-string',
            type=str,
            default=None,
            help='If set, this string is appended to all search queries.',
        )
        group.add_argument(
            '--krm-message-mutators',
            type=str,
            default=None,
            help='message mutators for the KRM model',
        )
        group.add_argument(
            '--drm-message-mutators',
            type=str,
            default=None,
            help='message mutators for the DRM model',
        )
        group.add_argument(
            '--search-server', type=str, default=None, help='search server to use.'
        )
        return parser

    def __init__(self, opt, shared=None):
        self.id = 'SeekerAgent'
        self.opt = opt
        one_model = opt.get('all_model_path') or opt.get('model_file')
        assert one_model, "Must specify a model file for this agent."
        for k in ['knowledge_', 'dialogue_', 'search_query_', 'search_decision_']:
            opt[f'{k}response_model_path'] = one_model
        if opt['search_server']:
            opt['krm_search_server'] = opt['search_server']
            if 'override' in opt:
                opt['override']['krm_search_server'] = opt['search_server']

        self._construct_subagent_opts(opt)
        self.search_query_agent = None
        self.search_decision_agent = None

        if not shared:
            self.knowledge_agent = create_agent(
                self.opts['knowledge_agent'], requireModelExists=True
            )
            logging.verbose("options for knowledge agent")
            if logging.logger.isEnabledFor(logging.VERBOSE):
                self.knowledge_agent.opt.log()
        else:
            self.knowledge_agent = create_agent_from_shared(
                shared['knowledge_agent_share']
            )
        self.knowledge_responses = ['__SILENCE__']

        # Monkey-patch 1: KRM Context Blocking (include knowledge responses)
        if opt['include_knowledge_in_krm_context_blocking']:
            orig_fun = self.knowledge_agent._get_batch_context
            if opt['exclude_context_in_krm_context_blocking']:
                self.knowledge_agent._get_batch_context = MethodType(
                    lambda self, batch: krm_get_batch_context_only_knowledge(
                        self, batch, orig_fun
                    ),
                    self.knowledge_agent,
                )
            else:
                self.knowledge_agent._get_batch_context = MethodType(
                    lambda self, batch: krm_get_batch_context(self, batch, orig_fun),
                    self.knowledge_agent,
                )

        # Monkey-patch 2: DRM Context Blocking (exclude knowledge responses)
        self.dialogue_agent = self._init_shared_model('dialogue_agent')
        if opt['beam_disregard_knowledge_for_context_blocking']:
            orig_fun = self.dialogue_agent._get_batch_context
            self.dialogue_agent._get_batch_context = MethodType(
                lambda self, batch: drm_get_batch_context(
                    self, batch, orig_fun=orig_fun
                ),
                self.dialogue_agent,
            )

        # Shared Agents
        self.search_query_agent = self._init_shared_model('search_query_agent')
        self.search_decision_agent = self._init_shared_model('search_decision_agent')

        self.dialogue_agent_clones = [self.dialogue_agent]
        if not shared:
            # the main agent needs batchsize-many clones for handling history during batch act.
            self.dialogue_agent_clones += [
                self.dialogue_agent.clone() for _ in range(opt.get('batchsize', 1) - 1)
            ]

        self._init_mutators(opt)

        # Other Attrs
        self.search_decision = SearchDecision(opt['search_decision'])
        self.min_knowledge_length_when_search = opt['min_knowledge_length_when_search']
        self.inject_query_string = opt.get('inject_query_string', '')

        super().__init__(opt, shared)

    @property
    def history(self) -> History:
        return self.dialogue_agent.history

    @property
    def knowledge_agent_history(self) -> History:
        return self.knowledge_agent.history

    def reset(self):
        """
        Reset all sub agents.
        """
        if self.knowledge_agent:
            self.knowledge_agent.reset()
        for agent in self.dialogue_agent_clones:
            agent.reset()
        if self.search_query_agent:
            self.search_query_agent.reset()
        if self.search_decision_agent:
            self.search_decision_agent.reset()
        self.knowledge_responses = ['__SILENCE__']

    def _init_mutators(self, opt: Opt):
        """
        Initialize mutator objects for sub agents.
        """
        self.krm_mutators = None
        self.drm_mutators = None
        if opt['krm_message_mutators']:
            logging.warning(
                'WARNING: If specifying KRM Mutators, they MUST be message mutators'
            )
            mutator_types = Mutator.load_mutator_types(opt.get('krm_message_mutators'))
            self.krm_mutators = [mutator(opt) for mutator in mutator_types]
        if opt['drm_message_mutators']:
            logging.warning(
                'WARNING: If specifying DRM Mutators, they MUST be message mutators'
            )
            mutator_types = Mutator.load_mutator_types(opt.get('drm_message_mutators'))
            self.drm_mutators = [mutator(opt) for mutator in mutator_types]

    def _init_shared_model(self, opt_key: str):
        """
        Initialize a shared version of the "knowledge" model.

        This just makes sure that each "agent" has the same params, but different
        history objects.

        :param opt_key:
            which sub agent to create with the shared model.
        """
        return super().init_shared_model(self.opts[opt_key], self.knowledge_agent)

    def _construct_subagent_opts(self, opt: Opt):
        """
        Construct opts for each sub agent.

        :param opt:
            original Opt.
        """
        self.opts = {}
        self.opts['init'] = opt
        override_opts = defaultdict(dict)
        agent_mapping = {
            'krm': 'knowledge_agent',
            'drm': 'dialogue_agent',
            'sqm': 'search_query_agent',
            'sdm': 'search_decision_agent',
        }
        for k, v in opt['override'].items():
            k_set = False
            for prefix, ag_opt in agent_mapping.items():
                if k.startswith(f'{prefix}_'):
                    override_opts[ag_opt][k] = v
                    k_set = True
                    break
            if not k_set:
                override_opts['general'][k] = v
        self.opts['override'] = override_opts
        for prefix, key in agent_mapping.items():
            if f'{prefix}_interactive_mode' not in override_opts[key]:
                override_opts[key][f'{prefix}_interactive_mode'] = opt.get(
                    'interactive_mode', False
                )

        for agent in agent_mapping.values():
            filename = opt[f"{agent.replace('agent', 'response')}_model_path"]
            if not filename:
                continue
            self.opts[agent] = self._get_subagent_opt(
                filename=filename,
                specific_override_args=override_opts[agent],
                general_override_args=override_opts['general'],
            )
            self.opts[agent]['model_file'] = filename
            self.opts[agent]['override']['model_file'] = filename

    def _get_subagent_opt(
        self,
        filename: str,
        specific_override_args: Dict[str, Any],
        general_override_args: Dict[str, Any],
    ) -> Opt:
        """
        Given an agent opt, construct the new opt for the agent.

        :param filename:
            opt path
        :param specific_override_args:
            args for the specific agent
        :param general_override_args:
            args specified for all agents
        """
        return super().get_subagent_opt(
            self.opt['datapath'],
            filename,
            specific_override_args,
            general_override_args,
        )

    def share(self):
        shared = super().share()
        shared['knowledge_agent_share'] = self.knowledge_agent.share()
        return shared

    def observe(self, observation: Message) -> Dict[str, Message]:
        """
        Observe in 3 out of the 4 modules.

        :param observation:
            incoming message

        :return self.observation:
            returned observation is actually a dictionary mapping
            agent module name to the corresponding observation
        """
        if not isinstance(observation, Message):
            observation = Message(observation)
        for key in ['label_candidates', 'knowledge']:
            # Delete unnecessarily large keys
            observation.pop(key, '')
        observation.force_set(
            'knowledge_response', observation.get('checked_sentence', '')
        )

        raw_observation = copy.deepcopy(observation)
        # This part is *specifically* for document chunking.
        if self.krm_mutators:
            observation = observation.copy()
            for mutator in self.krm_mutators:
                assert isinstance(mutator, MessageMutator), "not message mutator"
                observation = next(mutator([observation]))

        knowledge_observation = self.knowledge_agent.observe(observation)
        knowledge_observation['prior_knowledge_responses'] = ' '.join(
            self.knowledge_responses
        )
        if observation.get('episode_done'):
            self.knowledge_responses = ['__SILENCE__']
        search_query_observation = None
        if self.search_query_agent:
            sqm_obs = copy.deepcopy(observation)
            if self.opt['search_query_control_token']:
                sqm_obs.force_set(
                    'temp_history', f" {self.opt['search_query_control_token']}"
                )
            sqm_obs.force_set('skip_retrieval', True)
            search_query_observation = self.search_query_agent.observe(sqm_obs)

        search_decision_observation = None
        if (
            self.search_decision_agent
            and self.search_decision is SearchDecision.COMPUTE
        ):
            assert (
                self.search_decision_agent.history.size == 1
            ), "wrong history size! set --sdm-history-size 1"
            sdm_obs = copy.deepcopy(observation)
            if self.opt['search_decision_control_token']:
                sdm_obs.force_set(
                    'temp_history', f" {self.opt['search_decision_control_token']}"
                )
            sdm_obs.force_set('skip_retrieval', True)
            search_decision_observation = self.search_decision_agent.observe(sdm_obs)

        observations = {
            'raw': raw_observation,
            'knowledge_agent': knowledge_observation,
            'search_query_agent': search_query_observation,
            'search_decision_agent': search_decision_observation,
        }
        self.observations = observations
        return observations

    def batch_act_sdm(
        self,
        observations: List[Dict[str, Message]],
        knowledge_agent_observations: List[Message],
    ) -> Tuple[List[Message], List[int], List[Message]]:
        """
        Search Decision batch act.

        :param observations:
            observations for batch act.
        :param knowledge_agent_observations:
            observations to modify with the decision from the search decision agent.

        :return (batch_reply, search_indices, observations):
            batch_reply: reply from the search decision agent
            search_indices: batch indices with which to use search.
            observations: modified knowledge agent observations
        """
        search_indices = []
        batch_reply_sdm = [{} for _ in range(len(knowledge_agent_observations))]
        if self.search_decision is SearchDecision.ALWAYS:
            [o.force_set('skip_retrieval', False) for o in knowledge_agent_observations]
            search_indices = list(range(len(knowledge_agent_observations)))
        elif self.search_decision is SearchDecision.NEVER:
            [o.force_set('skip_retrieval', True) for o in knowledge_agent_observations]
        else:
            assert self.search_decision is SearchDecision.COMPUTE
            assert self.search_decision_agent
            batch_reply_sdm = self.search_decision_agent.batch_act(
                [o['search_decision_agent'] for o in observations]
            )
            for i, reply in enumerate(batch_reply_sdm):
                logging.debug(f"Example {i}: {reply['text']}")
                if reply['text'] == self.opt['search_decision_do_search_reply']:
                    search_indices.append(i)
                    knowledge_agent_observations[i].force_set('skip_retrieval', False)
                elif reply['text'] == self.opt['search_decision_dont_search_reply']:
                    knowledge_agent_observations[i].force_set('skip_retrieval', True)
                else:
                    logging.error(
                        f"SDM Reply: {reply['text']}; defaulting to no search"
                    )
                    knowledge_agent_observations[i].force_set('skip_retrieval', True)

        return batch_reply_sdm, search_indices, knowledge_agent_observations

    def batch_act_sqm(
        self, observations: List[Dict[str, Message]], search_indices: List[int]
    ) -> List[Message]:
        """
        Search Query Generator batch act.

        :param observations:
            list of observations
        :param search_indices:
            list of batch indices for which search is required.

        :return batch_reply:
            return the batch reply from the search query agent
        """
        return super().batch_act_search_query_generation(
            observations,
            [o['search_query_agent'] for o in observations],
            search_indices,
            self.search_query_agent,
            self.knowledge_agent,
            self.inject_query_string,
        )

    def batch_act_krm(
        self,
        observations: List[Dict[str, Message]],
        knowledge_agent_observations: List[Message],
        search_indices: List[int],
    ) -> List[Message]:
        """
        Knowledge Response Model batch act.

        :param observations:
            list of observations
        :param knowledge_agent_observations:
            observations for the knowledge agent.
        :param search_indices:
            list of indices for which we search.
            important for min length generation when search.

        :return batch_reply:
            batch_reply: batch reply from KRM
        """
        old_min_length = self.knowledge_agent.beam_min_length
        batch_reply_krm = [{} for _ in range(len(observations))]

        if search_indices and self.min_knowledge_length_when_search > 0:
            # Need to handle min length for searching
            self.knowledge_agent.beam_min_length = self.min_knowledge_length_when_search
            search_replies = self.knowledge_agent.batch_act(
                [
                    o
                    for i, o in enumerate(knowledge_agent_observations)
                    if i in search_indices
                ]
            )
            self.knowledge_agent.beam_min_length = old_min_length
            non_search_replies = (
                self.knowledge_agent.batch_act(
                    [
                        o
                        for i, o in enumerate(knowledge_agent_observations)
                        if i not in search_indices
                    ]
                )
                if len(search_indices) != len(observations)
                else []
            )
            search_offset = 0
            no_search_offset = 0
            for i in range(len(observations)):
                if i in search_indices:
                    batch_reply_krm[i] = search_replies[search_offset]
                    search_offset += 1
                else:
                    batch_reply_krm[i] = non_search_replies[no_search_offset]
                    no_search_offset += 1
        else:
            batch_reply_krm = self.knowledge_agent.batch_act(
                knowledge_agent_observations
            )

        self.knowledge_agent.beam_min_length = old_min_length

        return batch_reply_krm

    def batch_act_drm(
        self, observations: List[Dict[str, Message]], batch_reply_krm: List[Message]
    ) -> List[Message]:
        """
        Batch act in the DRM.

        Generate knowledge-infused observations (via temp history).
        Then batch act.
        Then observe all of the replies.

        :param observations:
            observations from self.observe
        :param batch_reply_krm:
            batch reply from the KRM module

        :return batch_reply_drm:
            return the reply from the DRM.
        """
        knowledge = [
            reply_knowledge.get('text', '') for reply_knowledge in batch_reply_krm
        ]
        full_text = [
            o['knowledge_agent'].get('full_text', o.get('text', ''))
            for o in observations
        ]
        logging.debug(f'Generated knowledge: {knowledge}')
        dialogue_agent_observations = []
        for i, obs in enumerate(observations):
            drm_obs = copy.deepcopy(obs['raw'])
            drm_obs.force_set(
                'temp_history',
                f"\n{TOKEN_KNOWLEDGE} {batch_reply_krm[i].get('text', '')} {TOKEN_END_KNOWLEDGE}",
            )
            drm_obs.force_set('skip_retrieval', True)
            if not self.dialogue_agent_clones[i].history.get_history_str():
                drm_obs.force_set('text', full_text[i])
            dialogue_agent_observations.append(
                self.dialogue_agent_clones[i].observe(drm_obs)
            )
            assert all(
                t in self.dialogue_agent_clones[i].history.get_history_str()
                for t in [TOKEN_KNOWLEDGE, TOKEN_END_KNOWLEDGE]
            )
        batch_reply_drm = self.dialogue_agent.batch_act(dialogue_agent_observations)
        for i, obs in enumerate(batch_reply_drm):
            self.dialogue_agent_clones[i].self_observe(obs)
            # manually clear
            self.dialogue_agent_clones[i].history.temp_history = None

        return batch_reply_drm

    def batch_act(self, observations: List[Dict[str, Message]]) -> List[Message]:
        """
        Full batch_act pipeline.

        :param observations:
            batchsize-length list of observations from self.observe

        :return reply:
            return batchsize-length list of final replies.
        """
        knowledge_agent_observations = [o['knowledge_agent'] for o in observations]
        # First, determine whether we're searching
        (
            batch_reply_sdm,
            search_indices,
            knowledge_agent_observations,
        ) = self.batch_act_sdm(observations, knowledge_agent_observations)
        # Second, generate search queries
        batch_reply_sqm = self.batch_act_sqm(observations, search_indices)

        # Third, generate the knowledge sentence
        batch_reply_krm = self.batch_act_krm(
            observations, knowledge_agent_observations, search_indices
        )

        # Fourth, generate the dialogue response!
        batch_reply_drm = self.batch_act_drm(observations, batch_reply_krm)

        # Finaly, combine them all in the drm batch reply.
        for sdm, sqm, krm, drm in zip(
            batch_reply_sdm, batch_reply_sqm, batch_reply_krm, batch_reply_drm
        ):
            if drm.is_padding():
                continue
            drm.force_set('search_decision', sdm.get('text', ''))
            drm.force_set('search_query', sqm.get('text', ''))
            drm.force_set('knowledge_response', krm.get('text', ''))
            docs = krm.get('top_docs', [Document("", "", "")])
            drm.force_set('doc_titles', [d.get_title() for d in docs])
            drm.force_set('doc_content', [d.get_text() for d in docs])
            drm.force_set('doc_urls', [d.get_id() for d in docs])

        return batch_reply_drm

    def act(self):
        """
        Call batch_act with the singleton batch.
        """
        response = self.batch_act([self.observations])[0]
        self.self_observe(response)
        return response

    def self_observe(self, self_message: Message):
        """
        Override TA.self_observe.

        Make sure that knowledge agent and other agents have the same history.

        This eliminates unnecessary copies of the previous knowledge in the history.
        """
        self.knowledge_agent.self_observe(self_message)
        self.knowledge_responses.append(self_message.get('knowledge_response', ''))
        observation = {'text': self.knowledge_agent.history.get_history_str()}
        for agent in self.dialogue_agent_clones:
            agent.reset()
            agent.history.update_history(
                observation,
                temp_history=self.dialogue_agent.get_temp_history(observation),
            )
        if (
            self.search_decision_agent
            and self.search_decision is SearchDecision.COMPUTE
        ):
            self.search_decision_agent.self_observe(self_message)
            self.search_decision_agent.history.reset()
            self.search_decision_agent.history.update_history(
                observation,
                temp_history=self.search_decision_agent.get_temp_history(observation),
            )
        if self.search_query_agent:
            self.search_query_agent.self_observe(self_message)
            self.search_query_agent.history.reset()
            self.search_query_agent.history.update_history(
                observation,
                temp_history=self.search_query_agent.get_temp_history(observation),
            )



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
"""
Base script for running TOD model-model chats.
"""

from parlai.core.params import ParlaiParser
from parlai.core.script import ParlaiScript, register_script

from projects.tod_simulator.scripts.get_passing_only import GetPassingOnlyScript

import glob


@register_script("do_get_passing_only_on_dir")
class DoGetPassingOnlyOnDirScript(ParlaiScript):
    @classmethod
    def setup_args(cls):
        parser = ParlaiParser(False, False)
        parser.add_argument("-p", "--path", required=True, type=str)
        parser.add_argument(
            "--filter-call-attempts",
            default=True,
            help="when True, only counts as 'passing' if System made exactly same # of api calls as goals",
        )
        return parser

    def run(self):
        opt = self.opt
        path = opt["path"]

        # assumes standard naming from the `model_consts` set of scripts
        base_paths = [
            x.replace("_conversations.jsonl", "")
            for x in glob.glob(f"{path}/*_conversations.jsonl")
        ]

        for to_run in base_paths:
            convo_path = to_run + "_conversations.jsonl"
            report_path = to_run + ".json"

            here_opt = {
                "convo_path": convo_path,
                "report_path": report_path,
                "print_to_file": True,
                "filter_call_attempts": opt["filter_call_attempts"],
            }
            GetPassingOnlyScript._run_kwargs(here_opt)
            print("done with ", convo_path)


if __name__ == "__main__":
    DoGetPassingOnlyOnDirScript.main()



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import os
from typing import List, Optional

import parlai.utils.logging as logging
from parlai.utils.io import PathManager
from parlai.core.message import Message
from parlai.core.params import ParlaiParser
from parlai.core.opt import Opt
from parlai.core.teachers import DialogTeacher
from .build import build


DEFAULT_TRAIN_EXPERIENCER_ONLY = False


class EmpatheticDialoguesTeacher(DialogTeacher):
    def __init__(self, opt, shared=None):
        build(opt)
        opt['datafile'] = self._get_datafile(opt)
        self.id = 'empathetic_dialogues'
        self.experiencer_side_only = self._get_experiencer_side_only(opt)
        super().__init__(opt, shared)

    @classmethod
    def add_cmdline_args(
        cls, parser: ParlaiParser, partial_opt: Optional[Opt] = None
    ) -> ParlaiParser:
        super().add_cmdline_args(parser, partial_opt)
        agent = parser.add_argument_group('EmpatheticDialogues teacher arguments')
        agent.add_argument(
            '--train-experiencer-only',
            type='bool',
            default=DEFAULT_TRAIN_EXPERIENCER_ONLY,
            # i.e. do not include the other side of the conversation where the Listener
            # (responder) utterance would be the text and the Speaker (experiencer)
            # utterance would be the label
            help='In the train set, only use Speaker (experiencer) utterances as text and Listener (responder) utterances as labels.',
        )
        return parser

    def _get_base_datatype(self, opt) -> str:
        return opt['datatype'].split(':')[0]

    def _get_datafile(self, opt) -> str:
        """
        Get the datafile path.

        Useful for subclassed teachers.
        """
        base_datatype = self._get_base_datatype(opt)
        return os.path.join(
            opt['datapath'],
            'empatheticdialogues',
            'empatheticdialogues',
            base_datatype + '.csv',
        )

    def _get_experiencer_side_only(self, opt):
        """
        Determine which side(s) of the conversation to use.
        """
        base_datatype = self._get_base_datatype(opt)
        return (
            opt.get('train_experiencer_only', DEFAULT_TRAIN_EXPERIENCER_ONLY)
            and base_datatype == 'train'
        ) or base_datatype != 'train'

    def setup_data(self, path):

        logging.debug('loading: ' + path)
        with PathManager.open(path) as f:
            df = f.readlines()

        turn_idx = 1
        responder_text_dialogue = []
        experiencer_text_dialogue = []
        data = []
        for i in range(1, len(df)):

            cparts = df[i - 1].strip().split(",")
            sparts = df[i].strip().split(",")

            if cparts[0] == sparts[0]:

                # Check that the turn number has incremented correctly
                turn_idx += 1
                assert (
                    int(cparts[1]) + 1 == int(sparts[1]) and int(sparts[1]) == turn_idx
                )

                contextt = cparts[5].replace("_comma_", ",")
                label = sparts[5].replace("_comma_", ",").strip()
                prompt = sparts[2]
                sit = sparts[3].replace("_comma_", ",")
                if len(sparts) == 9:
                    if sparts[8] != '':
                        inline_label_candidates = [
                            cand.replace("_comma_", ",").replace("_pipe_", "|")
                            for cand in sparts[8].split('|')
                        ]
                    else:
                        inline_label_candidates = None
                elif len(sparts) == 8:
                    inline_label_candidates = None
                else:
                    raise ValueError(f'Line {i:d} has the wrong number of fields!')

                dialogue_parts = Message(
                    {
                        'text': contextt,
                        'labels': [label],
                        'emotion': prompt,
                        'situation': sit,
                    }
                )
                if inline_label_candidates is not None:
                    inline_label_candidates = [
                        lc.strip() for lc in inline_label_candidates
                    ]
                    dialogue_parts.force_set(
                        'label_candidates', inline_label_candidates
                    )

                if int(sparts[1]) % 2 == 0:
                    # experiencer is the "text" and responder is the "label"
                    experiencer_text_dialogue.append(dialogue_parts)
                else:
                    # responder is the "text" and experiencer is the "label"
                    responder_text_dialogue.append(dialogue_parts)

            else:

                # We've finished the previous episode, so add it to the data
                turn_idx = 1
                data += self._select_dialogues_to_add(
                    experiencer_text_dialogue, responder_text_dialogue
                )
                experiencer_text_dialogue = []
                responder_text_dialogue = []

        # Add in the final episode
        data += self._select_dialogues_to_add(
            experiencer_text_dialogue, responder_text_dialogue
        )

        for episode in data:
            for entry_idx, entry in enumerate(episode):
                new_episode = entry_idx == 0
                yield entry, new_episode

    def _select_dialogues_to_add(
        self,
        experiencer_text_dialogue: List[Message],
        responder_text_dialogue: List[Message],
    ) -> List[List[Message]]:
        """
        Return conversation halves to add to self.data.

        Given lists corresponding to the conversation turns from both sides of the
        conversation, return only the list(s) that will be used by the teacher.
        Optionally include both sides of the conversation.
        """
        selected_dialogues = []
        if len(experiencer_text_dialogue) > 0:
            selected_dialogues.append(experiencer_text_dialogue)
        if len(responder_text_dialogue) > 0 and not self.experiencer_side_only:
            selected_dialogues.append(responder_text_dialogue)
        return selected_dialogues


class DefaultTeacher(EmpatheticDialoguesTeacher):
    pass
